import unittest

from tests.components.playground_table_clients import PlaygroundDynamoDBCachingTable
from tests.tests_float_serialization.table_models import DynamoDBTableModel


class TestsDynamoDBCachingTable(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.table_client = PlaygroundDynamoDBCachingTable(data_model=DynamoDBTableModel)
        self.table_client.debug = True

        self.SHARED_CASE_KWARGS = {
            'self': self, 'table_client': self.table_client,
            'is_caching': True, 'primary_key_name': 'accountId'
        }

    def test_get_field(self):
        from tests.tests_float_serialization.cases_shared import test_get_simple_float_field
        test_get_simple_float_field(**self.SHARED_CASE_KWARGS)

    def test_get_multiple_float_fields(self):
        from tests.tests_float_serialization.cases_shared import test_get_multiple_float_fields
        test_get_multiple_float_fields(**self.SHARED_CASE_KWARGS)

    def test_remove_simple_float_field(self):
        from tests.tests_float_serialization.cases_shared import test_remove_simple_float_field
        test_remove_simple_float_field(**self.SHARED_CASE_KWARGS)



import unittest
from typing import Optional, Union
from uuid import uuid4

from StructNoSQL import DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiCachingTable
from StructNoSQL.clients_middlewares.external_dynamodb_api.external_dynamodb_api_basic_table import ExternalDynamoDBApiBasicTable


def test_basic_record_removal(
        self: unittest.TestCase, users_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        primary_key_name: str, is_caching: bool
):
    random_record_id: str = f"recordId_{uuid4()}"
    random_value: str = f"value_{uuid4()}"
    put_record_success: bool = users_table.put_record(record_dict_data={
        primary_key_name: random_record_id, 'value': random_value
    })
    self.assertTrue(put_record_success)

    removed_record_data: Optional[dict] = users_table.remove_record(indexes_keys_selectors={primary_key_name: random_record_id})
    # todo: simplify the remove_record and delete_record, since deletion can only be done with the primary index,
    #  replace the (indexes_keys_selectors: dict) argument by a simple (primary_key_value: str)
    self.assertEqual((
        {primary_key_name: random_record_id, 'value': random_value} if is_caching is not True else
        {'value': {primary_key_name: random_record_id, 'value': random_value}, 'fromCache': False}
    ), removed_record_data)



import random
import unittest
from typing import Optional, Dict

from StructNoSQL import BaseField, MapModel, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    class NestedObjectItem(MapModel):
        value = BaseField(field_type=int, required=False)
    nestedObject = BaseField(field_type=Dict[str, NestedObjectItem], key_name='nestedObjectId', required=False)


class TestDatabaseFieldsInitialization(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_initialize_new_nested_object(self):
        field_random_value: int = random.randint(0, 100)
        project_id = "testFieldInitializationNewProjectId"
        query_kwargs = {'nestedObjectId': project_id}

        set_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='nestedObject.{{nestedObjectId}}.value',
            query_kwargs=query_kwargs, value_to_set=field_random_value
        )
        self.assertTrue(set_success)

        retrieved_value: Optional[int] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID, field_path='nestedObject.{{nestedObjectId}}.value', query_kwargs=query_kwargs
        )
        self.assertEqual(retrieved_value, field_random_value)

        deletion_success: bool = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='nestedObject.{{nestedObjectId}}',
            query_kwargs=query_kwargs
        )
        self.assertTrue(deletion_success)


if __name__ == '__main__':
    unittest.main()



from typing import Dict
from StructNoSQL import TableDataModel, BaseField, MapModel, NoneType


class UsersTableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    username = BaseField(field_type=str)
    class ProjectModel(MapModel):
        projectName = BaseField(field_type=str, required=True)
        class InstancesInfosModel(MapModel):
            ya = BaseField(field_type=str)
        instancesInfos = BaseField(field_type=InstancesInfosModel, required=False)
    projects = BaseField(field_type=Dict[str, ProjectModel], key_name="projectId")
    multiTypes = BaseField(field_type=[str, NoneType], required=True)
    number1 = BaseField(field_type=[int, float], required=False)
    string1 = BaseField(field_type=str, required=False)
    floatTest = BaseField(field_type=float, required=False)

    fieldToRemove = BaseField(field_type=str, required=False)
    fieldToDelete = BaseField(field_type=str, required=False)
    class SophisticatedRemovalModel(MapModel):
        nestedVariable = BaseField(field_type=str, required=False)
    sophisticatedRemoval = BaseField(field_type=Dict[str, SophisticatedRemovalModel], key_name="id", required=False)

    class TestMapModel(MapModel):
        sampleText = BaseField(field_type=str, required=False)
    testMapModel = BaseField(field_type=TestMapModel, required=False)

    testDictWithPrimitiveValue = BaseField(field_type=Dict[str, bool], key_name="key")



import unittest
from typing import Union, Optional
from uuid import uuid4

from StructNoSQL import DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiCachingTable, \
    ExternalDynamoDBApiBasicTable, FieldTargetNotFoundException
from tests.components.playground_table_clients import TEST_ACCOUNT_ID


def test_set_update_field_with_custom_name(
        self: unittest.TestCase, users_table: Union[DynamoDBBasicTable, DynamoDBCachingTable, ExternalDynamoDBApiBasicTable, ExternalDynamoDBApiCachingTable],
        primary_key_name: str, is_caching: bool
):
    simple_field_random_value = f"simpleField_{uuid4()}"
    simple_field_update_success: bool = users_table.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField', value_to_set=simple_field_random_value
    )
    self.assertTrue(simple_field_update_success)

    retrieved_simple_field_value: Optional[str] = users_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='simpleField'
    )
    self.assertEqual((
        simple_field_random_value if is_caching is not True else
        {'fromCache': True, 'value': simple_field_random_value}
    ), retrieved_simple_field_value)

    custom_name_field_random_value = f"customNameField_{uuid4()}"
    custom_name_field_update_success: bool = users_table.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='field%/!$', value_to_set=custom_name_field_random_value
    )
    self.assertTrue(custom_name_field_update_success)

    retrieved_custom_name_field_value: Optional[str] = users_table.get_field(
        key_value=TEST_ACCOUNT_ID, field_path='field%/!$'
    )
    self.assertEqual((
        custom_name_field_random_value if is_caching is not True else
        {'fromCache': True, 'value': custom_name_field_random_value}
    ), retrieved_custom_name_field_value)

    try:
        users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldWithCustomName')
        self.fail()
    except FieldTargetNotFoundException as e:
        pass

if __name__ == '__main__':
    unittest.main()



from StructNoSQL import TableDataModel, BaseField, MapModel


class BaseTableModel(TableDataModel):
    simpleFloatField = BaseField(field_type=float, required=False)
    class ContainerModel(MapModel):
        floatOne = BaseField(field_type=float, required=False)
        floatTwo = BaseField(field_type=float, required=False)
    floatsContainer = BaseField(field_type=ContainerModel, required=False)

class DynamoDBTableModel(BaseTableModel):
    accountId = BaseField(field_type=str, required=True)

class ExternalDynamoDBApiTableModel(BaseTableModel):
    accountProjectTableKeyId = BaseField(field_type=str, required=True)



import unittest
from typing import Dict, Any
from uuid import uuid4

from StructNoSQL import FieldSetter, DynamoDBCachingTable, DynamoDBBasicTable, FieldGetter
from tests.components.playground_table_clients import TEST_ACCOUNT_USERNAME, TEST_ACCOUNT_ID


def test_set_get_fields_with_secondary_index(self: unittest.TestCase, users_table: DynamoDBBasicTable or DynamoDBCachingTable, is_caching: bool):
    if is_caching is True:
        users_table.clear_cached_data_and_pending_operations()
        
    field_one_random_value = f"field1_{uuid4()}"
    field_two_random_value = f"field2_{uuid4()}"

    set_update_success: bool = users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
        FieldSetter(field_path='fieldOne', value_to_set=field_one_random_value),
        FieldSetter(field_path='fieldTwo', value_to_set=field_two_random_value)
    ])
    self.assertTrue(set_update_success)

    if is_caching is True:
        update_commit_success: bool = users_table.commit_operations()
        self.assertTrue(update_commit_success)

    retrieved_values = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='(fieldOne, fieldTwo)')
    self.assertEqual({
        'fieldOne': (
            field_one_random_value if is_caching is not True else
            {'fromCache': True, 'value': field_one_random_value}
        ),
        'fieldTwo': (
            field_two_random_value if is_caching is not True else
            {'fromCache': True, 'value': field_two_random_value}
        )
    }, retrieved_values)

    single_field_not_primary_key, query_metadata = users_table.query_field(key_value=TEST_ACCOUNT_USERNAME, index_name='username', field_path='fieldOne')
    self.assertEqual({TEST_ACCOUNT_ID: (
        field_one_random_value if is_caching is not True else
        {'fromCache': False, 'value': field_one_random_value}
    )}, single_field_not_primary_key)

    single_field_primary_key, query_metadata = users_table.query_field(key_value=TEST_ACCOUNT_USERNAME, index_name='username', field_path='accountId')
    self.assertEqual({TEST_ACCOUNT_ID: (
        TEST_ACCOUNT_ID if is_caching is not True else
        {'fromCache': False, 'value': TEST_ACCOUNT_ID}
    )}, single_field_primary_key)

    multiple_fields_without_primary_key, query_metadata = users_table.query_field(
        key_value=TEST_ACCOUNT_USERNAME, index_name='username', field_path='(fieldOne, fieldTwo)'
    )
    self.assertEqual({
        TEST_ACCOUNT_ID: {
            'fieldOne': (
                field_one_random_value if is_caching is not True else
                {'fromCache': False, 'value': field_one_random_value}
            ),
            'fieldTwo': (
                field_two_random_value if is_caching is not True else
                {'fromCache': False, 'value': field_two_random_value}
            )
        }}, multiple_fields_without_primary_key
    )

    multiple_fields_with_primary_key, query_metadata = users_table.query_field(
        key_value=TEST_ACCOUNT_USERNAME, index_name='username',
        field_path='(accountId, fieldOne, fieldTwo)'
    )
    self.assertEqual({
        TEST_ACCOUNT_ID: {
            'accountId': (
                TEST_ACCOUNT_ID if is_caching is not True else
                {'fromCache': False, 'value': TEST_ACCOUNT_ID}
            ),
            'fieldOne': (
                field_one_random_value if is_caching is not True else
                {'fromCache': False, 'value': field_one_random_value}
            ),
            'fieldTwo': (
                field_two_random_value if is_caching is not True else
                {'fromCache': False, 'value': field_two_random_value}
            )
        }}, multiple_fields_with_primary_key
    )

def test_set_get_fields_with_overriding_names(
        self: unittest.TestCase, users_table: DynamoDBBasicTable or DynamoDBCachingTable, is_caching: bool
):
    item1_field1_random_value: str = f"item1_field1_${uuid4()}"
    item2_field1_random_value: str = f"item2_field1_${uuid4()}"
    set_update_success: bool = users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
        FieldSetter(
            field_path='container.{{containerKey}}',
            query_kwargs={'containerKey': "item1"},
            value_to_set={'fieldOne': item1_field1_random_value}
        ),
        FieldSetter(
            field_path='container.{{containerKey}}',
            query_kwargs={'containerKey': "item2"},
            value_to_set={'fieldOne': item2_field1_random_value}
        )
    ])
    self.assertTrue(set_update_success)

    if is_caching is True:
        commit_success: bool = users_table.commit_operations()
        self.assertTrue(commit_success)

    retrieved_items_values, query_metadata = users_table.query_multiple_fields(
        key_value=TEST_ACCOUNT_USERNAME, index_name='username', getters={
            'item1-value': FieldGetter(field_path='container.{{containerKey}}.fieldOne', query_kwargs={'containerKey': "item1"}),
            'item2-value': FieldGetter(field_path='container.{{containerKey}}.fieldOne', query_kwargs={'containerKey': "item2"}),
        }
    )
    retrieved_items_values: Dict[str, Any]
    self.assertEqual({
        TEST_ACCOUNT_ID: {
            'item1-value': (
                item1_field1_random_value if is_caching is not True else
                {'fromCache': False, 'value': item1_field1_random_value}
            ),
            'item2-value': (
                item2_field1_random_value if is_caching is not True else
                {'fromCache': False, 'value': item2_field1_random_value}
            )
        }}, retrieved_items_values
    )

def test_set_get_fields_with_multi_selectors(
        self: unittest.TestCase, users_table: DynamoDBBasicTable or DynamoDBCachingTable, is_caching: bool
):
    field1_random_value: str = f"field1_${uuid4()}"
    field2_random_value: str = f"field2_${uuid4()}"
    set_update_success: bool = users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
        FieldSetter(field_path='fieldOne', value_to_set=field1_random_value),
        FieldSetter(field_path='fieldTwo', value_to_set=field2_random_value)
    ])
    self.assertTrue(set_update_success)

    if is_caching is True:
        commit_success: bool = users_table.commit_operations()
        self.assertTrue(commit_success)

    retrieved_items_values, query_metadata = users_table.query_field(
        key_value=TEST_ACCOUNT_USERNAME, index_name='username', field_path='(fieldOne, fieldTwo)'
    )
    retrieved_items_values: Dict[str, Any]
    self.assertEqual({
        TEST_ACCOUNT_ID: {
            'fieldOne': (
                field1_random_value if is_caching is not True else
                {'fromCache': False, 'value': field1_random_value}
            ),
            'fieldTwo': (
                field2_random_value if is_caching is not True else
                {'fromCache': False, 'value': field2_random_value}
            )
        }}, retrieved_items_values
    )

    retrieved_items_values, query_metadata = users_table.query_multiple_fields(
        key_value=TEST_ACCOUNT_USERNAME, index_name='username',
        getters={'item': FieldGetter(field_path='(fieldOne, fieldTwo)')}
    )
    retrieved_items_values: Dict[str, Any]
    print(retrieved_items_values)


def test_set_get_multiple_fields_with_special_inner_keys(
        self: unittest.TestCase, users_table: DynamoDBBasicTable or DynamoDBCachingTable, is_caching: bool
):
    """
    This test the client usage of '__PRIMARY_KEY__' as a getter key in a query_multiple_fields operation on a secondary index.
    The _inner_query_fields_secondary_index will be triggered to handle the requested, in our test case, we will not request
    the primary index, which will be added to the existing requested fields, by default with '__PRIMARY_KEY__' as the getter
    key, but since it is already used by the client, we should normally start to add random letters to '__PRIMARY_KEY__'
    until a free index key is found.
    """

    field1_random_value: str = f"field1_${uuid4()}"
    field2_random_value: str = f"field2_${uuid4()}"
    set_update_success: bool = users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
        FieldSetter(field_path='fieldOne', value_to_set=field1_random_value),
        FieldSetter(field_path='fieldTwo', value_to_set=field2_random_value)
    ])
    self.assertTrue(set_update_success)

    if is_caching is True:
        commit_success: bool = users_table.commit_operations()
        self.assertTrue(commit_success)

    retrieved_items_values, query_metadata = users_table.query_multiple_fields(
        key_value=TEST_ACCOUNT_USERNAME, index_name='username',
        getters={
            '__PRIMARY_KEY__': FieldGetter(field_path='fieldOne'),
            'fieldTwo': FieldGetter(field_path='fieldTwo'),
        }
    )
    self.assertEqual({
        TEST_ACCOUNT_ID: {
            '__PRIMARY_KEY__': (
                field1_random_value if is_caching is not True else
                {'fromCache': False, 'value': field1_random_value}
            ),
            'fieldTwo': (
                field2_random_value if is_caching is not True else
                {'fromCache': False, 'value': field2_random_value}
            )
        }}, retrieved_items_values
    )



from typing import Dict
from StructNoSQL import TableDataModel, BaseField, MapModel


class BaseTableModel(TableDataModel):
    value = BaseField(field_type=str, required=False)

class DynamoDBTableModel(BaseTableModel):
    accountId = BaseField(field_type=str, required=True)

class ExternalDynamoDBApiTableModel(BaseTableModel):
    accountProjectTableKeyId = BaseField(field_type=str, required=True)



import unittest
from typing import Optional, Dict
from uuid import uuid4

from StructNoSQL import BaseField, FieldRemover, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    rootFieldOne = BaseField(field_type=str, required=False)
    rootFieldTwo = BaseField(field_type=str, required=False)


class TestMultiRetrievalOfRootFilesWithMissingData(unittest.TestCase):
    """
    This test the main bug fix released in StructNoSQL 1.5.4, where if multiple root fields were requested
    (example, rootFieldOne and rootFieldTwo), if one of the field was found and got a value, but the other
    one was not found, and so not included in the response_data dict, we would end up with a response data
    with only key. Prior to 1.5.4, StructNoSQL would treat any returned dictionary with a single key as a 
    DynamoDB data container, and would try to convert it. Usually, a wrong conversion like this would not
    cause any problems, unless there was another field we requested, in which case, the returned data would
    be navigated into too much by StructNoSQL, and would then be used to populate the multiple requested
    fields (ie, populate the rootFieldOne value in both rootFieldOne and rootFieldTwo, where rootFieldTwo
    should be None). StructNoSQL 1.5.4 should have fixed this issue by not treating any dict with a single
    key as a DynamoDB container, but only dict with a single key that is a DynamoDB data types keys (ie: 'S',
    'BOOL', 'NULL', etc), and by now considering all the DynamoDB data types keys as restricted fields names.
    """

    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_with_two_root_fields(self):
        deletions_successes: Dict[str, bool] = self.users_table.delete_multiple_fields(
            key_value=TEST_ACCOUNT_ID, removers={
                'one': FieldRemover(field_path='rootFieldOne'),
                'two': FieldRemover(field_path='rootFieldTwo')
            }
        )

        field_one_random_value = f"field1_{uuid4()}"
        field_one_update_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path='rootFieldOne', value_to_set=field_one_random_value
        )
        self.assertTrue(field_one_update_success)

        retrieved_data: Optional[dict] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID, field_path='(rootFieldOne, rootFieldTwo)'
        )
        self.assertIsNotNone(retrieved_data)
        self.assertEqual(retrieved_data.get('rootFieldOne', None), field_one_random_value)
        self.assertEqual(retrieved_data.get('rootFieldTwo', None), None)


if __name__ == '__main__':
    unittest.main()



from typing import List
from StructNoSQL import TableDataModel, BaseField, MapModel


class BaseTableModel(TableDataModel):
    simpleValue = BaseField(field_type=int, required=False)
    simpleValue2 = BaseField(field_type=int, required=False)
    fieldOne = BaseField(field_type=str, required=False)
    fieldTwo = BaseField(field_type=str, required=False)
    fieldToDelete = BaseField(field_type=int, required=False)
    fieldToDelete2 = BaseField(field_type=int, required=False)
    fieldToRemove = BaseField(field_type=int, required=False)
    fieldToRemove2 = BaseField(field_type=int, required=False)
    class ContainerToRemoveModel(MapModel):
        fieldOne = BaseField(field_type=str, required=False)
        fieldTwo = BaseField(field_type=str, required=False)
        fieldThree = BaseField(field_type=str, required=False)
    containerToRemove = BaseField(field_type=ContainerToRemoveModel, required=False)
    containersListToRemove = BaseField(field_type=List[ContainerToRemoveModel], key_name='listIndex')

class DynamoDBTableModel(BaseTableModel):
    accountId = BaseField(field_type=str, required=True)

class ExternalDynamoDBApiTableModel(BaseTableModel):
    accountProjectTableKeyId = BaseField(field_type=str, required=True)



import random
import unittest
from typing import List, Callable, Dict, Optional
from uuid import uuid4
from StructNoSQL import FieldRemover, FieldSetter, FieldGetter
from StructNoSQL.clients_middlewares.dynamodb.dynamodb_caching_table import DynamoDBCachingTable
from StructNoSQL.clients_middlewares.external_dynamodb_api.external_dynamodb_api_caching_table import ExternalDynamoDBApiCachingTable
from tests.components.playground_table_clients import TEST_ACCOUNT_ID, TEST_ACCOUNT_USERNAME


# todo: add an unit test that make sure that what matter with the field are the field names, not their variable names


def test_simple_get_field(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()

    first_response_data = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='simpleValue')
    self.assertEqual(first_response_data['fromCache'], False)

    second_response_data = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='simpleValue')
    self.assertEqual(second_response_data['fromCache'], True)

def test_set_then_get_field_from_cache(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()
    random_field_value = random.randint(0, 100)

    update_success = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='simpleValue', value_to_set=random_field_value)
    self.assertTrue(update_success)

    retrieve_response_data = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='simpleValue')
    self.assertEqual(retrieve_response_data['fromCache'], True)

def test_set_then_get_multiple_fields(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()

    random_field_one_value = random.randint(0, 99)
    random_field_two_value = random.randint(100, 199)

    update_success = users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
        FieldSetter(field_path='simpleValue', value_to_set=random_field_one_value),
        FieldSetter(field_path='simpleValue2', value_to_set=random_field_two_value)
    ])
    self.assertTrue(update_success)

    retrieve_response_data = users_table.get_multiple_fields(key_value=TEST_ACCOUNT_ID, getters={
        'one': FieldGetter(field_path='simpleValue'),
        'two': FieldGetter(field_path='simpleValue2')
    })
    self.assertEqual(retrieve_response_data.get('one', None), {'value': random_field_one_value, 'fromCache': True})
    self.assertEqual(retrieve_response_data.get('two', None), {'value': random_field_two_value, 'fromCache': True})

def test_set_then_get_pack_values_with_one_of_them_present_in_cache(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()

    random_field_one_value = random.randint(0, 99)
    random_field_two_value = random.randint(100, 199)

    update_success = users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
        FieldSetter(field_path='simpleValue', value_to_set=random_field_one_value),
        FieldSetter(field_path='simpleValue2', value_to_set=random_field_two_value)
    ])
    self.assertTrue(update_success)

    users_table.commit_operations()
    users_table.clear_cached_data_and_pending_operations()

    # Caching the simpleValue field
    first_retrieved_first_value = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='simpleValue')
    self.assertEqual({'value': random_field_one_value, 'fromCache': False}, first_retrieved_first_value)

    # With get_field function and multi selector
    get_field_response_data = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='(simpleValue, simpleValue2)')
    self.assertEqual(get_field_response_data.get('simpleValue', None), {'value': random_field_one_value, 'fromCache': True})
    self.assertEqual(get_field_response_data.get('simpleValue2', None), {'value': random_field_two_value, 'fromCache': False})

    users_table.clear_cached_data_and_pending_operations()
    # Caching the simpleValue field
    second_retrieved_first_value = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='simpleValue')
    self.assertEqual(second_retrieved_first_value, {'fromCache': False, 'value': random_field_one_value})

    # With get_multiple_fields function
    get_multiple_fields_response_data = users_table.get_multiple_fields(
        key_value=TEST_ACCOUNT_ID, getters={
            'one': FieldGetter(field_path='simpleValue'),
            'two': FieldGetter(field_path='simpleValue2')
        }
    )
    self.assertEqual(get_multiple_fields_response_data.get('one', None), {'value': random_field_one_value, 'fromCache': True})
    self.assertEqual(get_multiple_fields_response_data.get('two', None), {'value': random_field_two_value, 'fromCache': False})

def test_debug_simple_set_commit_then_get_field_from_database(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()
    random_field_value = random.randint(0, 100)

    update_success = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='simpleValue', value_to_set=random_field_value)
    self.assertTrue(update_success)
    commit_success = users_table.commit_operations()
    self.assertTrue(commit_success)

    users_table.clear_cached_data_and_pending_operations()
    retrieve_response_data = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='simpleValue')
    self.assertEqual(retrieve_response_data, {'fromCache': False, 'value': random_field_value})

def test_update_multiple_fields(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()
    random_field_one_value = random.randint(0, 99)
    random_field_two_value = random.randint(100, 199)

    update_success = users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
        FieldSetter(field_path='simpleValue', value_to_set=random_field_one_value),
        FieldSetter(field_path='simpleValue2', value_to_set=random_field_two_value)
    ])
    self.assertTrue(update_success)

    retrieved_data = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='(simpleValue, simpleValue2)')
    self.assertEqual(retrieved_data.get('simpleValue', None), {'value': random_field_one_value, 'fromCache': True})
    self.assertEqual(retrieved_data.get('simpleValue2', None), {'value': random_field_two_value, 'fromCache': True})

def test_set_delete_field(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()
    random_field_value = random.randint(0, 100)

    update_success = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete', value_to_set=random_field_value)
    self.assertTrue(update_success)
    update_commit_success = users_table.commit_operations()
    self.assertTrue(update_commit_success)

    delete_success = users_table.delete_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete')
    self.assertTrue(delete_success)
    delete_commit_success = users_table.commit_operations()
    self.assertTrue(delete_commit_success)

    retrieved_expected_empty_value_from_cache = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete')
    self.assertEqual(retrieved_expected_empty_value_from_cache, {'fromCache': True, 'value': None})

    users_table.clear_cached_data_and_pending_operations()
    retrieved_expected_empty_value_from_database = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete')
    self.assertEqual(retrieved_expected_empty_value_from_database, {'fromCache': False, 'value': None})

def test_set_remove_field(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()
    random_field_value = random.randint(0, 100)

    update_success = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove', value_to_set=random_field_value)
    self.assertTrue(update_success)
    update_commit_success = users_table.commit_operations()
    self.assertTrue(update_commit_success)

    removed_value = users_table.remove_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove')
    self.assertEqual(removed_value, {'fromCache': True, 'value': random_field_value})

    retrieved_expected_empty_value_from_cache = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove')
    self.assertEqual(retrieved_expected_empty_value_from_cache, {'fromCache': True, 'value': None})

    users_table.commit_operations()
    users_table.clear_cached_data_and_pending_operations()

    retrieved_expected_empty_value_from_database = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove')
    self.assertEqual(retrieved_expected_empty_value_from_database, {'fromCache': False, 'value': None})

def test_dict_data_unpacking(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    """
    Data unpacking handle the scenario where an object would be put in the cache (like a dictionary, that has been
    updated as an object, instead of doing it field per field), and then later on, we try to access from the cache
    one field of packed data. We want to be handle to retrieve data that was packed in bigger object, which means
    that we cannot use a simple flatten indexation of the inserted/updated data to then later access it in the cache.
    """
    users_table.clear_cached_data_and_pending_operations()
    random_field_one_value = f"field_one_{uuid4()}"
    random_field_two_value = f"field_two_{uuid4()}"

    update_success = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='containerToRemove', value_to_set={
        'fieldOne': random_field_one_value, 'fieldTwo': random_field_two_value
    })
    self.assertTrue(update_success)
    update_commit_success = users_table.commit_operations()
    self.assertTrue(update_commit_success)

    retrieved_value = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='containerToRemove.(fieldOne, fieldTwo)')
    self.assertEqual(retrieved_value.get('fieldOne', {}), {'value': random_field_one_value, 'fromCache': True})
    self.assertEqual(retrieved_value.get('fieldTwo', {}), {'value': random_field_two_value, 'fromCache': True})

    removed_value = users_table.remove_field(key_value=TEST_ACCOUNT_ID, field_path='containerToRemove.(fieldOne, fieldTwo)')
    self.assertEqual(removed_value.get('fieldOne', {}), {'value': random_field_one_value, 'fromCache': True})
    self.assertEqual(removed_value.get('fieldTwo', {}), {'value': random_field_two_value, 'fromCache': True})

def test_list_data_unpacking(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    """
    Data unpacking handle the scenario where an object would be put in the cache (like a dictionary, that has been
    updated as an object, instead of doing it field per field), and then later on, we try to access from the cache
    one field of packed data. We want to be handle to retrieve data that was packed in bigger object, which means
    that we cannot use a simple flatten indexation of the inserted/updated data to then later access it in the cache.
    """
    users_table.clear_cached_data_and_pending_operations()
    random_field_one_value = f"field_one_{uuid4()}"
    random_field_two_value = f"field_two_{uuid4()}"

    update_success = users_table.update_field(
        key_value=TEST_ACCOUNT_ID, field_path='containersListToRemove.{{listIndex}}', query_kwargs={'listIndex': 0},
        value_to_set={'fieldOne': random_field_one_value, 'fieldTwo': random_field_two_value}
    )
    self.assertTrue(update_success)
    update_commit_success = users_table.commit_operations()
    self.assertTrue(update_commit_success)

    retrieved_value = users_table.get_field(
        key_value=TEST_ACCOUNT_ID, query_kwargs={'listIndex': 0},
        field_path='containersListToRemove.{{listIndex}}.(fieldOne, fieldTwo)'
    )
    self.assertEqual(retrieved_value.get('fieldOne', {}), {'value': random_field_one_value, 'fromCache': True})
    self.assertEqual(retrieved_value.get('fieldTwo', {}), {'value': random_field_two_value, 'fromCache': True})

    removed_values = users_table.remove_field(
        key_value=TEST_ACCOUNT_ID, query_kwargs={'listIndex': 0},
        field_path='containersListToRemove.{{listIndex}}.(fieldOne, fieldTwo)'
    )
    self.assertEqual(removed_values.get('fieldOne', {}), {'value': random_field_one_value, 'fromCache': True})
    self.assertEqual(removed_values.get('fieldTwo', {}), {'value': random_field_two_value, 'fromCache': True})

def test_set_remove_multi_selector_field_and_field_unpacking(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()
    random_field_one_value = f"field_one_{uuid4()}"
    random_field_two_value = f"field_two_{uuid4()}"
    random_field_three_value = f"field_three_{uuid4()}"

    update_success = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='containerToRemove', value_to_set={
        'fieldOne': random_field_one_value, 'fieldTwo': random_field_two_value, 'fieldThree': random_field_three_value
    })
    self.assertTrue(update_success)
    update_commit_success = users_table.commit_operations()
    self.assertTrue(update_commit_success)

    removed_value = users_table.remove_field(key_value=TEST_ACCOUNT_ID, field_path='containerToRemove.(fieldOne, fieldThree)')
    self.assertEqual(removed_value.get('fieldOne', {}), {'fromCache': True, 'value': random_field_one_value})
    self.assertEqual(removed_value.get('fieldThree', {}), {'fromCache': True, 'value': random_field_three_value})

    users_table.clear_cached_data_and_pending_operations()
    removed_values: Optional[dict] = users_table.remove_field(key_value=TEST_ACCOUNT_ID, field_path='containerToRemove.(fieldOne, fieldThree)')
    self.assertEqual(removed_values, {
        'fieldOne': {'fromCache': False, 'value': random_field_one_value},
        'fieldThree': {'fromCache': False, 'value': random_field_three_value}
    })

def test_set_delete_multiple_fields(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()
    random_field_value_one = random.randint(0, 100)
    random_field_value_two = random.randint(100, 200)

    update_success_one = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete', value_to_set=random_field_value_one)
    self.assertTrue(update_success_one)
    update_success_two = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete2', value_to_set=random_field_value_two)
    self.assertTrue(update_success_two)
    update_commit_success = users_table.commit_operations()
    self.assertTrue(update_commit_success)

    retrieved_data_one = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete')
    self.assertEqual(retrieved_data_one['value'], random_field_value_one)
    retrieved_data_two = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete2')
    self.assertEqual(retrieved_data_two['value'], random_field_value_two)

    multi_delete_response = users_table.delete_multiple_fields(
        key_value=TEST_ACCOUNT_ID, removers={
            'one': FieldRemover(field_path='fieldToDelete'),
            'two': FieldRemover(field_path='fieldToDelete2')
        }
    )
    self.assertTrue(all(multi_delete_response.values()))

    retrieved_expected_empty_value_one_from_cache = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete')
    self.assertTrue(retrieved_expected_empty_value_one_from_cache['fromCache'])
    self.assertIsNone(retrieved_expected_empty_value_one_from_cache['value'])
    retrieved_expected_empty_value_two_from_cache = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete2')
    self.assertTrue(retrieved_expected_empty_value_two_from_cache['fromCache'])
    self.assertIsNone(retrieved_expected_empty_value_two_from_cache['value'])

    users_table.commit_operations()
    users_table.clear_cached_data_and_pending_operations()

    retrieved_expected_empty_value_one_from_database = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete')
    self.assertFalse(retrieved_expected_empty_value_one_from_database['fromCache'])
    self.assertIsNone(retrieved_expected_empty_value_one_from_database['value'])
    retrieved_expected_empty_value_two_from_database = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToDelete2')
    self.assertFalse(retrieved_expected_empty_value_two_from_database['fromCache'])
    self.assertIsNone(retrieved_expected_empty_value_two_from_database['value'])

def test_set_remove_multiple_fields(self: unittest.TestCase, users_table: DynamoDBCachingTable or ExternalDynamoDBApiCachingTable):
    users_table.clear_cached_data_and_pending_operations()
    random_field_value_one = random.randint(0, 100)
    random_field_value_two = random.randint(100, 200)

    update_success_one = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove', value_to_set=random_field_value_one)
    self.assertTrue(update_success_one)
    update_success_two = users_table.update_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove2', value_to_set=random_field_value_two)
    self.assertTrue(update_success_two)
    update_commit_success = users_table.commit_operations()
    self.assertTrue(update_commit_success)

    retrieved_data_one = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove')
    self.assertEqual(retrieved_data_one['value'], random_field_value_one)
    retrieved_data_two = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove2')
    self.assertEqual(retrieved_data_two['value'], random_field_value_two)

    response_data = users_table.remove_multiple_fields(key_value=TEST_ACCOUNT_ID, removers={
        'one': FieldRemover(field_path='fieldToRemove'),
        'two': FieldRemover(field_path='fieldToRemove2')
    })
    self.assertEqual({'fromCache': True, 'value': random_field_value_one}, response_data.get('one', None))
    self.assertEqual({'fromCache': True, 'value': random_field_value_two}, response_data.get('two', None))

    retrieved_expected_empty_value_one_from_cache = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove')
    self.assertTrue(retrieved_expected_empty_value_one_from_cache['fromCache'])
    self.assertIsNone(retrieved_expected_empty_value_one_from_cache['value'])
    retrieved_expected_empty_value_two_from_cache = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove2')
    self.assertTrue(retrieved_expected_empty_value_two_from_cache['fromCache'])
    self.assertIsNone(retrieved_expected_empty_value_two_from_cache['value'])

    users_table.commit_operations()
    users_table.clear_cached_data_and_pending_operations()

    retrieved_expected_empty_value_one_from_database = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove')
    self.assertFalse(retrieved_expected_empty_value_one_from_database['fromCache'])
    self.assertIsNone(retrieved_expected_empty_value_one_from_database['value'])
    retrieved_expected_empty_value_two_from_database = users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove2')
    self.assertFalse(retrieved_expected_empty_value_two_from_database['fromCache'])
    self.assertIsNone(retrieved_expected_empty_value_two_from_database['value'])



import re
from typing import List, Optional, Any, Dict, _GenericAlias, Tuple

from StructNoSQL.utils.objects import NoneType
from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_utils import DynamoDBUtils
from StructNoSQL.models import DatabasePathElement
from StructNoSQL.exceptions import InvalidFieldNameException, UsageOfUntypedSetException
from StructNoSQL.practical_logger import message_with_vars
from StructNoSQL.query import Query
from StructNoSQL.utils.types import ACCEPTABLE_KEY_TYPES
from StructNoSQL.utils.misc_fields_items import make_dict_key_var_name, try_to_get_primitive_default_type_of_item


FIELD_NAME_RESTRICTED_CHARS_LIST = ['[', ']', '{', '}', '(', ')', '|']
FIELD_NAME_RESTRICTED_CHARS_EXPRESSION = "|".join(re.escape(char) for char in FIELD_NAME_RESTRICTED_CHARS_LIST)
# We escape each individual char, which will add the appropriate backslashes to the chars that require them.

FIELD_NAME_RESTRICTED_NAMES_LIST = DynamoDBUtils.DYNAMODB_TYPES_KEYS
FIELD_NAME_RESTRICTED_NAMES_EXPRESSION = f"^({'|'.join(name for name in FIELD_NAME_RESTRICTED_NAMES_LIST)})$"


def _raise_if_field_name_is_invalid(field_name: str):
    restricted_chars_matches: Optional[List[tuple]] = re.findall(pattern=FIELD_NAME_RESTRICTED_CHARS_EXPRESSION, string=field_name)
    if restricted_chars_matches is not None and len(restricted_chars_matches) > 0:
        raise InvalidFieldNameException(message_with_vars(
            message="A field name was using one or multiple restricted chars",
            vars_dict={
                'fieldName': field_name, 'restrictedCharsMatch': restricted_chars_matches,
                'FIELD_NAME_RESTRICTED_CHARS_LIST': FIELD_NAME_RESTRICTED_CHARS_LIST
            }
        ))

    restricted_names_matches: Optional[List[tuple]] = re.findall(pattern=FIELD_NAME_RESTRICTED_NAMES_EXPRESSION, string=field_name)
    if restricted_names_matches is not None and len(restricted_names_matches) > 0:
        raise InvalidFieldNameException(message_with_vars(
            message="A field name was using a restricted name",
            vars_dict={
                'fieldName': field_name, 'restricted_names_matches': restricted_names_matches,
                'FIELD_NAME_RESTRICTED_NAMES_LIST': FIELD_NAME_RESTRICTED_NAMES_LIST
            }
        ))


class BaseDataModel:
    def __init__(self):
        self.childrens_map: Optional[dict] = None

class MapModel(BaseDataModel):
    _default_primitive_type = dict
    required_fields: list = None

    def __init__(self, **kwargs):
        super().__init__()
        self.kwargs = kwargs
        # from StructNoSQL import field_loader
        # field_loader.load(class_instance=self, **kwargs)

    def new(self):
        pass

    @property
    def dict(self) -> dict:
        return self.childrens_map if self.childrens_map is not None else self.kwargs


class TableDataModel(MapModel):
    # The TableDataModel inherit from MapModel, to allow easier validation of record data.
    # For example, when the put_record function is used, and needs data validation.
    def class_add_field(self, field_key: str, field_item: Any):
        setattr(self.__class__, field_key, field_item)

    def class_from_fields(self, fields: Dict[str, Any]):
        for key, item in fields.items():
            self.class_add_field(field_key=key, field_item=item)

    def instance_add_field(self, field_key: str, field_item: Any):
        self.__dict__[field_key] = field_item

    def instance_from_fields(self, fields: Dict[str, Any]):
        for key, item in fields.items():
            self.instance_add_field(field_key=key, field_item=item)


def _alias_to_model(alias: _GenericAlias):
    alias_variable_name: Optional[str] = alias.__dict__.get('_name', None)
    if alias_variable_name is not None:
        alias_args: Optional[Tuple] = alias.__dict__.get('__args__', None)
        if alias_args is not None:
            if alias_variable_name == "Dict":
                return DictModel(key_type=alias_args[0], item_type=alias_args[1])
            elif alias_variable_name == "Set":
                raise Exception("not yet implemented")
            elif alias_variable_name == "List":
                raise Exception("not yet implemented")
    return None


class BaseItem:
    _table = None
    _database_path: Optional[List[DatabasePathElement]] = None
    _key_expected_type: Optional[type] = None
    _items_excepted_type: Optional[type or MapModel] = None
    # We set the _database_path as static, so that the assign_internal_mapping_from_class can setup the path only once,
    # only by having access to the inheritor class type, not even the instance. Yet, when set the _database_path
    # statically, the value is not attributed to the BaseField class (which would cause to have multiple classes override
    # the paths of the others), but rather, it is statically set on the class that inherit from BaseField.

    def __init__(self, field_type: Optional[type] = Any, custom_default_value: Optional[Any] = None):
        # todo: add a file_url field_type
        self._value = None
        self._query = None

        self.map_key_expected_type: Optional[type] = None
        self.map_model: Optional[MapModel or type] = None

        self._custom_default_value = custom_default_value
        self._field_type = field_type
        self._default_field_type = field_type

        if isinstance(self._field_type, (list, tuple)):
            if not len(self._field_type) > 0:
                raise Exception("At least one field_type must be specified in a list or tuple of field types")
            self._default_field_type = self._field_type[0]

        elif isinstance(self._field_type, _GenericAlias):
            alias_variable_name: Optional[str] = self._field_type.__dict__.get('_name', None)
            if alias_variable_name is not None:
                alias_args: Optional[Tuple] = self._field_type.__dict__.get('__args__', None)

                if alias_variable_name == "Dict":
                    self._field_type = dict
                    self._default_field_type = dict
                    self._key_expected_type = alias_args[0]
                    if self._key_expected_type not in ACCEPTABLE_KEY_TYPES:
                        raise Exception(message_with_vars(
                            message="Key in a Dict field was not found in the acceptable key types",
                            vars_dict={'_key_expected_type': self._key_expected_type, 'ACCEPTABLE_KEY_TYPES': ACCEPTABLE_KEY_TYPES}
                        ))

                    self._items_excepted_type = alias_args[1]
                    if isinstance(self._items_excepted_type, _GenericAlias):
                        self._items_excepted_type = _alias_to_model(alias=self._items_excepted_type)

                elif alias_variable_name == "Set":
                    self._field_type = set
                    self._default_field_type = set
                    self._items_excepted_type = alias_args[0]
                    if isinstance(self._items_excepted_type, _GenericAlias):
                        self._items_excepted_type = _alias_to_model(alias=self._items_excepted_type)
                    # todo: rename the _items_excepted_type variable
                elif alias_variable_name == "List":
                    self._field_type = list
                    self._default_field_type = list
                    self._items_excepted_type = alias_args[0]
                    if isinstance(self._items_excepted_type, _GenericAlias):
                        self._items_excepted_type = _alias_to_model(alias=self._items_excepted_type)
                    # todo: rename the _items_excepted_type variable
                else:
                    raise Exception(f"Unsupported GenericAlias : {alias_variable_name}")

        elif MapModel in getattr(self._field_type, '__mro__', ()):
            self.map_model = self._field_type
            self._field_type = dict
            self._default_field_type = dict

        elif self._field_type == dict:
            # Handle an untyped dict
            self._items_excepted_type = Any
            self._key_expected_type = Any
        elif self._field_type == list:
            # Handle an untyped list
            self._items_excepted_type = Any
        elif self._field_type == set:
            # Raise on an untyped set
            raise UsageOfUntypedSetException()

    def validate_data(self) -> Tuple[Optional[Any], bool]:
        from StructNoSQL.validator import validate_data
        validated_data, valid = validate_data(
            value=self._value, item_type_to_return_to=self,
            expected_value_type=self._field_type,
        )
        self._value = validated_data
        return validated_data, valid

    def populate(self, value: Any):
        self._value = value
        # self.validate_data()
        # print("Finished data validation.")

    def query(self, key_value: str, fields_path_elements: List[str], index_name: Optional[str] = None, query_kwargs: Optional[dict] = None) -> Query:
        self._query = Query(
            variable_validator=self, table=self._table,
            target_database_path=self._database_path,
            index_name=index_name, key_value=key_value,
        )
        return self._query

    def get_default_value(self):
        if self._custom_default_value is not None:
            return self._custom_default_value
        # We initialize the default field type
        # and return it as the default value.
        return self._field_type()

    @property
    def custom_default_value(self) -> Optional[Any]:
        return self._custom_default_value

    @property
    def value(self) -> Any:
        return self._value

    @property
    def field_type(self) -> Any:
        return self._field_type

    @property
    def default_field_type(self) -> type:
        return self._default_field_type

    @property
    def key_expected_type(self) -> Optional[type]:
        return self._key_expected_type

    @property
    def items_excepted_type(self) -> Optional[type or MapModel]:
        return self._items_excepted_type

    @property
    def database_path(self) -> Optional[List[DatabasePathElement]]:
        return self._database_path

    @property
    def table(self):
        return self._table

    @staticmethod
    def instantiate_default_value_type(value_type: type) -> Optional[Any]:
        if value_type == Any:
            return None
        else:
            try:
                return value_type()
            except Exception as e:
                print(e)
                return None


class BaseField(BaseItem):
    def __init__(
            self, field_type: Any, required: Optional[bool] = False, not_modifiable: Optional[bool] = False, custom_field_name: Optional[str] = None,
            custom_default_value: Optional[Any] = None, key_name: Optional[str] = None, max_nested_depth: Optional[int] = 32
    ):
        super().__init__(field_type=field_type, custom_default_value=custom_default_value)

        self._field_name = custom_field_name
        if self._field_name is not None:
            _raise_if_field_name_is_invalid(field_name=self._field_name)
        # If a custom_field_name has been specified by the user, it will be set as the field_name
        # (which will then be validated for invalid characters). The field_name will not be able to be
        # modified a second time (otherwise the field_name property setter will cause an exception),
        # so the process_item function of the base_table.py will see that the field_name has already been
        # initialized, and will not try to initialize it from the variable key_name from the class signature.

        self._required = required
        self._key_name = None

        if self._required is not True:
            # If the field is not required, we add the NoneType to the field_type
            if isinstance(self._field_type, list):
                # If it's a list, we just a need to append the NoneType
                self._field_type.append(NoneType)
            elif isinstance(self._field_type, tuple):
                # If it's a tuple, we combine the existing tuple with a tuple containing the NoneType
                self._field_type = self._field_type + (NoneType,)
            else:
                # And it it's any other type, we create a new tuple with the field_type and the NoneType
                self._field_type = (self._field_type, NoneType)

        if max_nested_depth is not None and max_nested_depth > 32:
            raise Exception(f"DynamoDB support a maximum depth of nested of items of 32. This is not imposed by StructNoSQL but a platform limitation.\n"
                            f"See : https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Limits.html#limits-attributes")
        self._max_nested = max_nested_depth

        if not_modifiable is True:
            raise Exception(f"Not modifiable not yet implemented")

        if key_name is not None:
            if field_type in [dict, set, list] or type(field_type) is _GenericAlias:
                self._key_name = key_name
            elif isinstance(field_type, (tuple, list)):
                raise Exception(f"Multiple dictionaries are not yet supported.")
                all_items_are_dict = True
                for item in field_type:
                    item_type = type(item)
                    if not isinstance(item, (dict, _GenericAlias)):
                        all_items_are_dict = False
                        break
                if all_items_are_dict is True:
                    self._key_name = key_name
                else:
                    raise Exception(message_with_vars(
                        message="key_name cannot be set on a field that is a tuple or list that does not exclusively contains dict or Dict items",
                        vars_dict={'fieldName': name, 'fieldType': field_type, 'keyName': key_name}
                    ))
            else:
                raise Exception(message_with_vars(
                    message="key_name cannot be set on a field that is not of type dict, Dict, list, List, set or Set",
                    vars_dict={'fieldName': self.field_name, 'fieldType': field_type, 'keyName': key_name}
                ))

    def populate(self, value: any):
        super().populate(value=value)

    @property
    def dict_item(self):
        """ :return: BaseField """
        if self._field_type == dict and self.items_excepted_type is not None:
            map_item = MapItem(parent_field=self, field_type=self.items_excepted_type, model_type=self.items_excepted_type)
            return map_item
        else:
            raise Exception(message_with_vars(
                message="Tried to access dict item of a field that was not of type dict, "
                        "Dict or did not properly received the expected type of the dict items.",
                vars_dict={"fieldName": self.field_name, "fieldType": self.field_type,
                           "dictItemsExceptedType": self.items_excepted_type}
            ))

    @property
    def field_name(self) -> str:
        return self._field_name

    @field_name.setter
    def field_name(self, value: str):
        if self._field_name is not None:
            raise Exception("field_name already defined")
        _raise_if_field_name_is_invalid(field_name=value)
        self._field_name = value

        # When field_name is set, if the field_type is one of the field_type's requiring a key_name and the
        # key_name has not be set by the user, we create a default key_name by adding 'Key' to the field_name.
        if self.field_type in [dict, set] or type(self.field_type) == _GenericAlias:
            if self._key_name is None:
                self._key_name = f"{self.field_name}Key"

    @property
    def key_name(self) -> Optional[str]:
        return self._key_name

    @property
    def required(self) -> bool:
        return self._required

    @property
    def max_nested(self) -> int:
        # todo: find a better name than max_nested
        return self._max_nested

    def copy(self):
        from copy import copy
        return copy(self)

class MapItem(BaseField):
    _default_primitive_type = dict

    def __init__(self, parent_field: BaseField, model_type: type, field_type: type):
        super().__init__(field_type=field_type, custom_default_value=BaseItem.instantiate_default_value_type(field_type))
        self.map_model = model_type

        self._key_name = parent_field.key_name
        # The _key_name will be initialized by the call to the super().__init__, but since a MapItem is special and does
        # not include a name, the key_name that the BaseField will create will be invalid. For MapItem's, we just want
        # to re-use the key_name of the parent_field (which will have been already prepared and processed in cases of
        # nested recursive fields), and override the existing _key_name value.

        element_key = make_dict_key_var_name(parent_field.key_name)
        default_type = try_to_get_primitive_default_type_of_item(parent_field.items_excepted_type)
        database_path_element = DatabasePathElement(element_key=element_key, default_type=default_type)
        self._database_path = [*parent_field.database_path, database_path_element]
        self._table = parent_field.table

class DictModel(BaseItem):
    _default_primitive_type = dict

    def __init__(self, key_type: Any, item_type: Any, key_name: Optional[str] = None):
        super().__init__()
        self.key_type = key_type
        self.item_type = item_type
        self.key_name = key_name
        self._database_path = None

    @property
    def database_path(self):
        return self._database_path



import random
import time
decs = ["*", "@", "~"]
while True:
    for i in range(11):
        dec_left = random.choice(decs)
        dec_right = random.choice(decs)
        spaces_left = " " * (11 - (i - 1))
        inside_tree = "~" * (i * 2)
        print(f"{spaces_left}{dec_left}</{inside_tree}/>{dec_right}")
    for i in range(3):
        print(f"{' ' * 11}{'#' * 6}")
    print("\n\n\n\n\n\n\n\n\n\n")
    time.sleep(0.5)



import abc
from typing import Optional, List, Dict, Any, Tuple, Callable, Union, Type

from StructNoSQL import BaseField, TableDataModel
from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_core import PrimaryIndex
from StructNoSQL.models import DatabasePathElement, FieldGetter, FieldRemover, FieldSetter, UnsafeFieldSetter, \
    FieldPathSetter, QueryMetadata
from StructNoSQL.practical_logger import message_with_vars
from StructNoSQL.tables.base_table import BaseTable
from StructNoSQL.tables.shared_table_behaviors import _model_contain_all_index_keys, \
    unpack_validate_retrieved_field_if_need_to, unpack_validate_multiple_retrieved_fields_if_need_to
from StructNoSQL.utils.data_processing import navigate_into_data_with_field_path_elements
from StructNoSQL.utils.process_render_fields_paths import process_and_make_single_rendered_database_path, \
    process_validate_data_and_make_single_rendered_database_path, join_field_path_elements


class BaseCachingTable(BaseTable):
    def __init__(self, data_model: Type[TableDataModel], primary_index: PrimaryIndex):
        super().__init__(data_model=data_model, primary_index=primary_index)
        self._cached_data_per_primary_key: Dict[str, Any] = {}
        self._pending_update_operations_per_primary_key: Dict[str, Dict[str, FieldPathSetter]] = {}
        self._pending_remove_operations_per_primary_key: Dict[str, Dict[str, List[DatabasePathElement]]] = {}
        self._debug = False
        self._wrap_item_value = BaseCachingTable._without_debug_wrap_item_value

    @staticmethod
    def _without_debug_wrap_item_value(item_value: Any, from_cache: bool) -> Any:
        return item_value

    @staticmethod
    def _with_debug_wrap_item_value(item_value: Any, from_cache: bool) -> Any:
        return {'value': item_value, 'fromCache': from_cache}

    @property
    def wrap_item_value(self):
        return self._wrap_item_value

    def wrap_item_value_from_cache(self, item_value: Any) -> Any:
        return self._wrap_item_value(item_value=item_value, from_cache=True)

    def wrap_item_value_not_from_cache(self, item_value: Any) -> Any:
        return self._wrap_item_value(item_value=item_value, from_cache=False)

    @property
    def debug(self) -> bool:
        return self._debug

    @debug.setter
    def debug(self, debug: bool):
        self._debug = debug
        self._wrap_item_value = (
            BaseCachingTable._with_debug_wrap_item_value
            if debug is True else
            BaseCachingTable._without_debug_wrap_item_value
        )

    def clear_cached_data(self):
        self._cached_data_per_primary_key = {}

    def clear_cached_data_for_record(self, record_primary_key: str):
        self._cached_data_per_primary_key.pop(record_primary_key, None)

    def clear_pending_update_operations(self):
        self._pending_update_operations_per_primary_key = {}

    def clear_pending_remove_operations(self):
        self._pending_remove_operations_per_primary_key = {}

    def clear_pending_operations(self):
        self.clear_pending_update_operations()
        self.clear_pending_remove_operations()

    def clear_cached_data_and_pending_operations(self):
        self.clear_cached_data()
        self.clear_pending_operations()

    def _index_cached_data(self, primary_key_value: str) -> dict:
        if primary_key_value not in self._cached_data_per_primary_key:
            self._cached_data_per_primary_key[primary_key_value] = {}
        return self._cached_data_per_primary_key[primary_key_value]

    def _remove_index_from_cached_data(self, primary_key_value: str) -> Optional[dict]:
        return self._cached_data_per_primary_key.pop(primary_key_value, None)

    def _index_pending_update_operations(self, primary_key_value: str) -> dict:
        if primary_key_value not in self._pending_update_operations_per_primary_key:
            self._pending_update_operations_per_primary_key[primary_key_value] = {}
        return self._pending_update_operations_per_primary_key[primary_key_value]

    def _index_pending_remove_operations(self, primary_key_value: str) -> dict:
        if primary_key_value not in self._pending_remove_operations_per_primary_key:
            self._pending_remove_operations_per_primary_key[primary_key_value] = {}
        return self._pending_remove_operations_per_primary_key[primary_key_value]

    def has_pending_update_operations(self) -> bool:
        for index_operations in self._pending_update_operations_per_primary_key.values():
            if len(index_operations) > 0:
                return True
        return False

    def has_pending_remove_operations(self) -> bool:
        for index_operations in self._pending_remove_operations_per_primary_key.values():
            if len(index_operations) > 0:
                return True
        return False

    def has_pending_operations(self) -> bool:
        return self.has_pending_update_operations() or self.has_pending_remove_operations()

    @abc.abstractmethod
    def commit_update_operations(self) -> bool:
        raise Exception("commit_update_operations not implemented")

    @abc.abstractmethod
    def commit_remove_operations(self) -> bool:
        raise Exception("commit_remove_operations not implemented")

    def commit_operations(self) -> bool:
        update_operations_commit_success: bool = self.commit_update_operations()
        remove_operations_commit_success: bool = self.commit_remove_operations()
        return all([update_operations_commit_success, remove_operations_commit_success])

    @staticmethod
    def _cache_put_data(index_cached_data: dict, field_path_elements: List[DatabasePathElement], data: Any):
        if len(field_path_elements) > 0:
            navigated_cached_data = index_cached_data
            for path_element in field_path_elements[:-1]:
                stringed_element_key = f'{path_element.element_key}'
                # We wrap the element_key inside a string, to handle a scenario where we would put an item from a list,
                # where the element_key will be an int, that could be above zero, and cannot be handled by a classical list.

                if stringed_element_key not in navigated_cached_data:
                    navigated_cached_data[stringed_element_key] = {}
                navigated_cached_data = navigated_cached_data[stringed_element_key]

            last_field_path_element = field_path_elements[-1]
            navigated_cached_data[f'{last_field_path_element.element_key}'] = data
            # todo: handle list's and set's

    def _cache_get_data(self, primary_key_value: str, field_path_elements: List[DatabasePathElement]) -> Tuple[bool, Any]:
        if len(field_path_elements) > 0:
            index_cached_data: dict = self._index_cached_data(primary_key_value=primary_key_value)
            navigated_cached_data: dict = index_cached_data

            for path_element in field_path_elements[:-1]:
                stringed_element_key = f'{path_element.element_key}'
                # We wrap the element_key inside a string, to handle a scenario where we would put an item from a list,
                # where the element_key will be an int, that could be above zero, and cannot be handled by a classical list.
                navigated_cached_data = (
                    navigated_cached_data[stringed_element_key]
                    if stringed_element_key in navigated_cached_data
                    else path_element.get_default_value()
                )

            last_field_path_element: DatabasePathElement = field_path_elements[-1]
            if last_field_path_element.element_key in navigated_cached_data:
                retrieved_item_value: Any = navigated_cached_data[last_field_path_element.element_key]
                return True, retrieved_item_value
        return False, None

    def _cache_delete_field(self, index_cached_data: dict, primary_key_value: str, field_path_elements: List[DatabasePathElement]) -> str:
        """Will remove the element value from the cache, and remove any update operations associated with the same field in the same index and key_value"""
        BaseCachingTable._cache_put_data(index_cached_data=index_cached_data, field_path_elements=field_path_elements, data=None)

        pending_update_operations: dict = self._index_pending_update_operations(primary_key_value=primary_key_value)
        item_joined_field_path: str = join_field_path_elements(field_path_elements)

        if item_joined_field_path in pending_update_operations:
            pending_update_operations.pop(item_joined_field_path)
        return item_joined_field_path

    def _cache_remove_field(self, index_cached_data: dict, primary_key_value: str, field_path_elements: List[DatabasePathElement]):
        """Unlike the _cache_delete_field, this must be used when a remove operation to the database will be performed right away"""
        item_joined_field_path = self._cache_delete_field(
            index_cached_data=index_cached_data,
            primary_key_value=primary_key_value,
            field_path_elements=field_path_elements
        )
        pending_remove_operations = self._index_pending_remove_operations(primary_key_value=primary_key_value)
        if item_joined_field_path in pending_remove_operations:
            pending_remove_operations.pop(item_joined_field_path)

    def _cache_add_delete_operation(self, index_cached_data: dict, pending_remove_operations: dict, field_path_elements: List[DatabasePathElement]):
        self._cache_put_data(index_cached_data=index_cached_data, field_path_elements=field_path_elements, data=None)
        # We put a None value in the cache, instead of removing it. Because if we de delete an item, we have the
        # in-memory knowledge that its expected value will be None, we do not need to forget its in-memory representation.
        joined_field_path = join_field_path_elements(field_path_elements)
        pending_remove_operations[joined_field_path] = field_path_elements

    def _cache_process_add_delete_operation(self, index_cached_data: dict, pending_remove_operations: dict, field_path: str, query_kwargs: Optional[dict] = None):
        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs
        )
        if is_multi_selector is not True:
            target_field_container: Tuple[BaseField, List[DatabasePathElement]]
            field_object, field_path_elements = target_field_container

            self._cache_add_delete_operation(
                index_cached_data=index_cached_data,
                pending_remove_operations=pending_remove_operations,
                field_path_elements=field_path_elements
            )
        else:
            target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

            for item_field_container in target_field_container.values():
                item_field_object, item_field_path_elements = item_field_container

                self._cache_add_delete_operation(
                    index_cached_data=index_cached_data,
                    pending_remove_operations=pending_remove_operations,
                    field_path_elements=item_field_path_elements
                )

    def _validate_cache_format_field_value_if_need_to(
            self, value: Any, data_validation: bool,
            field_object: BaseField, field_path_elements: List[DatabasePathElement],
            index_cached_data: dict, from_cache: bool
    ) -> Optional[Any]:
        if data_validation is True:
            field_object.populate(value=value)
            validated_data, is_valid = field_object.validate_data()
        else:
            validated_data = value

        BaseCachingTable._cache_put_data(
            index_cached_data=index_cached_data,
            field_path_elements=field_path_elements,
            data=validated_data
        )
        return self.wrap_item_value(item_value=validated_data, from_cache=from_cache)

    def _validate_cache_format_fields_values_if_need_to(
            self, data_validation: bool, record_item_data: dict, primary_key_value: str,
            target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]
    ) -> dict:
        output: dict = {}
        index_cached_data: dict = self._index_cached_data(primary_key_value=primary_key_value)

        for item_key, item_container in target_fields_containers.items():
            item_field_object, item_field_path_elements = item_container

            record_item_matching_item_data: Optional[Any] = record_item_data.get(item_key, None)
            output[item_key] = self._validate_cache_format_field_value_if_need_to(
                value=record_item_matching_item_data, data_validation=data_validation,
                field_object=item_field_object, field_path_elements=item_field_path_elements,
                index_cached_data=index_cached_data, from_cache=False
            )
        return output

    @staticmethod
    def _validate_field_value_if_need_to(value: Any, data_validation: bool, field_object: BaseField) -> Optional[Any]:
        if data_validation is not True:
            return value
        else:
            field_object.populate(value=value)
            validated_data, is_valid = field_object.validate_data()
            # Even if is_valid is False, we still return the validated_data, which will be None.
            return validated_data

    def _validate_format_field_value_if_need_to(self, value: Any, data_validation: bool, field_object: BaseField, from_cache: bool) -> Optional[Any]:
        validated_data: Optional[Any] = BaseCachingTable._validate_field_value_if_need_to(
            value=value, data_validation=data_validation, field_object=field_object
        )
        return self.wrap_item_value(item_value=validated_data, from_cache=from_cache)

    def _put_record(self, middleware: Callable[[dict], bool], record_dict_data: dict) -> bool:
        self.model_virtual_map_field.populate(value=record_dict_data)
        validated_data, is_valid = self.model_virtual_map_field.validate_data()
        put_record_success: bool = middleware(validated_data) if is_valid is True else False
        if put_record_success is True:
            self._cached_data_per_primary_key[record_dict_data[self.primary_index_name]] = record_dict_data
        return put_record_success

    def _delete_record(self, middleware: Callable[[dict], bool], indexes_keys_selectors: dict) -> bool:
        found_all_indexes: bool = _model_contain_all_index_keys(model=self.model, indexes_keys=indexes_keys_selectors.keys())
        if found_all_indexes is not True:
            return False

        deletion_success: bool = middleware(indexes_keys_selectors)
        if deletion_success is True:
            self._remove_index_from_cached_data(primary_key_value=indexes_keys_selectors[self.primary_index_name])
        return deletion_success

    def _remove_record(
            self, middleware: Callable[[dict], Optional[dict]],
            indexes_keys_selectors: dict, data_validation: bool
    ) -> Optional[dict]:
        found_all_indexes: bool = _model_contain_all_index_keys(model=self.model, indexes_keys=indexes_keys_selectors.keys())
        if found_all_indexes is not True:
            return None

        removed_record_data: Optional[dict] = middleware(indexes_keys_selectors)
        if removed_record_data is None:
            return None

        self._remove_index_from_cached_data(primary_key_value=indexes_keys_selectors[self.primary_index_name])
        return self._validate_format_field_value_if_need_to(
            value=removed_record_data, data_validation=data_validation,
            field_object=self.model_virtual_map_field, from_cache=False
        )

    def _get_field(
            self, middleware: Callable[[Union[List[DatabasePathElement], Dict[str, List[DatabasePathElement]]], bool], Any],
            key_value: str, field_path: str, query_kwargs: Optional[dict], data_validation: bool
    ) -> Optional[Any]:

        index_cached_data: dict = self._index_cached_data(primary_key_value=key_value)
        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs
        )
        if is_multi_selector is not True:
            target_field_container: Tuple[BaseField, List[DatabasePathElement]]
            field_object, field_path_elements = target_field_container

            found_in_cache, field_value_from_cache = self._cache_get_data(
                primary_key_value=key_value, field_path_elements=field_path_elements
            )
            if found_in_cache is True:
                return self._validate_format_field_value_if_need_to(
                    value=field_value_from_cache, data_validation=data_validation,
                    field_object=field_object, from_cache=True
                )

            retrieved_data: Optional[Any] = middleware(field_path_elements, False)
            return self._validate_cache_format_field_value_if_need_to(
                value=retrieved_data, data_validation=data_validation,
                field_object=field_object, field_path_elements=field_path_elements,
                index_cached_data=index_cached_data, from_cache=False
            )
        else:
            target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

            output_items: Dict[str, Optional[Any]] = {}
            keys_fields_already_cached_to_pop: List[str] = []

            for item_key, item_container in target_field_container.items():
                item_field_object, item_field_path_elements = item_container

                found_item_value_in_cache, field_item_value_from_cache = self._cache_get_data(
                    primary_key_value=key_value, field_path_elements=item_field_path_elements
                )
                if found_item_value_in_cache is True:
                    output_items[item_key] = self._validate_format_field_value_if_need_to(
                        value=field_item_value_from_cache, data_validation=data_validation,
                        field_object=item_field_object, from_cache=True
                    )
                    keys_fields_already_cached_to_pop.append(item_key)

            for key_to_pop in keys_fields_already_cached_to_pop:
                target_field_container.pop(key_to_pop)

            if len(target_field_container) > 0:
                fields_paths_elements: Dict[str, List[DatabasePathElement]] = {key: item[1] for key, item in target_field_container.items()}
                retrieved_items_data: Dict[str, Optional[Any]] = middleware(fields_paths_elements, True)
                for item_key, item_container in target_field_container.items():
                    item_field_object, item_field_path_elements = item_container
                    matching_item_data: Optional[Any] = retrieved_items_data.get(item_key, None)

                    output_items[item_key] = self._validate_cache_format_field_value_if_need_to(
                        value=matching_item_data, data_validation=data_validation,
                        field_object=item_field_object, field_path_elements=item_field_path_elements,
                        index_cached_data=index_cached_data, from_cache=False
                    )
            return output_items

    def inner_query_fields_secondary_index(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Tuple[Optional[List[Any]], QueryMetadata]],
            fields_database_paths: List[List[DatabasePathElement]],
    ) -> Tuple[Optional[dict], QueryMetadata]:
        from StructNoSQL.tables.shared_table_behaviors import _inner_query_fields_secondary_index
        return _inner_query_fields_secondary_index(
            primary_index_name=self.primary_index_name,
            get_primary_key_database_path=self._get_primary_key_database_path,
            middleware=middleware,
            fields_paths_elements=fields_database_paths
        )

    def format_item_value_if_need_to(self, item_value: Any):
        pass

    def unpack_validate_cache_getters_record_attributes_if_need_to(
            self, single_getters_target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]],
            grouped_getters_target_fields_containers: Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]],
            data_validation: bool, record_attributes: dict, index_cached_data: dict, base_output_values: Optional[Dict[str, Any]] = None
    ):
        def item_mutator(item_value: Any, item_field_path_elements: List[DatabasePathElement]) -> Any:
            BaseCachingTable._cache_put_data(
                index_cached_data=index_cached_data,
                field_path_elements=item_field_path_elements,
                data=item_value
            )
            return self.wrap_item_value(item_value=item_value, from_cache=False)

        from StructNoSQL.tables.shared_table_behaviors import _unpack_validate_getters_record_attributes_if_need_to
        return _unpack_validate_getters_record_attributes_if_need_to(
            single_getters_target_fields_containers=single_getters_target_fields_containers,
            grouped_getters_target_fields_containers=grouped_getters_target_fields_containers,
            item_mutator=item_mutator, data_validation=data_validation,
            record_attributes=record_attributes, base_output_values=base_output_values
        )

    def _shared_rar(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Tuple[List[Any], QueryMetadata]],
            key_value: str, target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]], data_validation: bool
    ) -> Tuple[Optional[dict], QueryMetadata]:
        existing_record_data: dict = {}

        keys_fields_already_cached_to_pop: List[str] = []
        for item_key, item_container in target_fields_containers.items():
            item_field_object, item_field_path_elements = item_container
            found_item_value_in_cache, field_item_value_from_cache = self._cache_get_data(
                primary_key_value=key_value, field_path_elements=item_field_path_elements
            )
            if found_item_value_in_cache is True:
                existing_record_data[item_key] = self._validate_format_field_value_if_need_to(
                    value=field_item_value_from_cache, data_validation=data_validation,
                    field_object=item_field_object, from_cache=True
                )
                keys_fields_already_cached_to_pop.append(item_key)

        for key_to_pop in keys_fields_already_cached_to_pop:
            target_fields_containers.pop(key_to_pop)

        if not len(target_fields_containers) > 0:
            return {key_value: existing_record_data}, QueryMetadata(count=1, has_reached_end=True, last_evaluated_key=None)

        fields_paths_elements: List[List[DatabasePathElement]] = [item[1] for item in target_fields_containers.values()]
        retrieved_records_items_data, query_metadata = middleware(fields_paths_elements)
        if retrieved_records_items_data is not None and len(retrieved_records_items_data) > 0:
            # Since we query the primary_index, we know for a fact that we will never be returned more than
            # one record item. Hence why we do not have a loop that iterate over the records_items_data,
            # and that we return a dict with the only one record item being the requested key_value.
            index_cached_data: dict = self._index_cached_data(primary_key_value=key_value)
            def item_mutator(item_value: Any, item_field_path_elements: List[DatabasePathElement]) -> Any:
                BaseCachingTable._cache_put_data(
                    index_cached_data=index_cached_data,
                    field_path_elements=item_field_path_elements,
                    data=item_value
                )
                return self.wrap_item_value(item_value=item_value, from_cache=False)

            rar_item_value: Optional[Any] = unpack_validate_multiple_retrieved_fields_if_need_to(
                target_fields_containers=target_fields_containers,
                data_validation=data_validation,
                record_attributes=retrieved_records_items_data[0],
                item_mutator=item_mutator,
                base_output_values=existing_record_data
            )
            # todo: rename rar_item_value
            return {key_value: rar_item_value}, query_metadata

    def _query_field(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Tuple[Optional[List[Any]], QueryMetadata]],
            key_value: str, field_path: str, query_kwargs: Optional[dict], index_name: Optional[str],
            data_validation: bool
    ) -> Tuple[Optional[dict], QueryMetadata]:

        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs
        )
        if index_name is not None and index_name != self.primary_index_name:
            output_records_values: dict = {}

            if is_multi_selector is not True:
                target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                fields_database_paths: List[List[DatabasePathElement]] = [target_field_container[1]]

                records_attributes, query_metadata = self.inner_query_fields_secondary_index(
                    middleware=middleware, fields_database_paths=fields_database_paths
                )
                for record_key, record_attributes in records_attributes.items():
                    index_cached_data: dict = self._index_cached_data(primary_key_value=record_key)
                    def item_mutator(item_value: Any, item_field_path_elements: List[DatabasePathElement]) -> Any:
                        BaseCachingTable._cache_put_data(
                            index_cached_data=index_cached_data,
                            field_path_elements=item_field_path_elements,
                            data=item_value
                        )
                        return self.wrap_item_value(item_value=item_value, from_cache=False)

                    output_records_values[record_key] = unpack_validate_retrieved_field_if_need_to(
                        record_attributes=record_attributes,
                        target_field_container=target_field_container,
                        data_validation=data_validation,
                        item_mutator=item_mutator
                    )
            else:
                target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]
                fields_database_paths: List[List[DatabasePathElement]] = [item[1] for item in target_field_container.values()]

                records_attributes, query_metadata = self.inner_query_fields_secondary_index(
                    middleware=middleware, fields_database_paths=fields_database_paths
                )
                for record_key, record_item_attributes in records_attributes.items():
                    index_cached_data: dict = self._index_cached_data(primary_key_value=record_key)
                    def item_mutator(item_value: Any, item_field_path_elements: List[DatabasePathElement]) -> Any:
                        BaseCachingTable._cache_put_data(
                            index_cached_data=index_cached_data,
                            field_path_elements=item_field_path_elements,
                            data=item_value
                        )
                        return self.wrap_item_value(item_value=item_value, from_cache=False)

                    output_records_values[record_key] = unpack_validate_multiple_retrieved_fields_if_need_to(
                        target_fields_containers=target_field_container,
                        record_attributes=record_item_attributes,
                        data_validation=data_validation,
                        item_mutator=item_mutator,
                    )
            return output_records_values, query_metadata
        else:
            # If requested index is primary index
            if is_multi_selector is not True:
                target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                field_object, field_path_elements = target_field_container

                found_in_cache, field_value_from_cache = self._cache_get_data(
                    primary_key_value=key_value, field_path_elements=field_path_elements
                )
                if found_in_cache is True:
                    # A single field was requested from a primary index, which means that only one record will have been 
                    # found for the specified index (primary index values are unique). Since the query_field function needs 
                    # to return a Dict[str, Any] with keys being the primary_key_value of the record, the the item being 
                    # the request data, we need to create a dict that wrap our retrieved value as a record result.
                    query_metadata = QueryMetadata(count=1, has_reached_end=True, last_evaluated_key=None)
                    # We can also create a representation of what the query metadata should look like, all from the in memory data.

                    output_item_value = self.wrap_item_value(item_value=field_value_from_cache, from_cache=True)
                    return {key_value: output_item_value}, query_metadata

                records_attributes, query_metadata = middleware([field_path_elements])
                if records_attributes is not None and len(records_attributes) > 0:
                    # Since we query the primary_index, we know for a fact that we will never be returned more than
                    # one record item. Hence why we do not have a loop that iterate over the records_items_data,
                    # and that we return a dict with the only one record item being the requested key_value.
                    single_record_attributes: dict = records_attributes[0]
                    navigated_record_item_value: Optional[Any] = navigate_into_data_with_field_path_elements(
                        data=single_record_attributes,
                        field_path_elements=field_path_elements,
                        num_keys_to_navigation_into=len(field_path_elements)
                    )
                    index_cached_data: dict = self._index_cached_data(primary_key_value=key_value)
                    validated_record_item_value: Optional[Any] = self._validate_cache_format_field_value_if_need_to(
                        value=navigated_record_item_value, data_validation=data_validation,
                        field_object=field_object, field_path_elements=field_path_elements,
                        index_cached_data=index_cached_data, from_cache=False
                    )
                    return {key_value: validated_record_item_value}, query_metadata
                return None, query_metadata
            else:
                target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]
                return self._shared_rar(
                    middleware=middleware, key_value=key_value,
                    target_fields_containers=target_field_container,
                    data_validation=data_validation
                )

    def _prepare_getters_with_cache(self, data_validation: bool, key_value: str, getters: Dict[str, FieldGetter]) -> Tuple[
        dict,
        List[List[DatabasePathElement]],
        Dict[str, Tuple[BaseField, List[DatabasePathElement]]],
        Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]]
    ]:
        existing_values: dict = {}
        getters_database_paths: List[List[DatabasePathElement]] = []
        single_getters_target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
        grouped_getters_target_fields_containers: Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]] = {}

        for getter_key, getter_item in getters.items():
            target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
                field_path=getter_item.field_path, fields_switch=self.fields_switch, query_kwargs=getter_item.query_kwargs
            )
            if is_multi_selector is not True:
                target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                field_object, field_path_elements = target_field_container
                found_item_value_in_cache, field_item_value_from_cache = self._cache_get_data(
                    primary_key_value=key_value, field_path_elements=field_path_elements
                )
                if found_item_value_in_cache is True:
                    if data_validation is True:
                        field_object.populate(value=field_item_value_from_cache)
                        validated_data, is_valid = field_object.validate_data()
                    else:
                        validated_data = field_item_value_from_cache
                    existing_values[getter_key] = self.wrap_item_value(item_value=validated_data, from_cache=True)
                else:
                    single_getters_target_fields_containers[getter_key] = target_field_container
                    getters_database_paths.append(target_field_container[1])
            else:
                target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

                container_existing_values: dict = {}
                container_fields: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}

                field_path_object: Dict[str, BaseField]
                getter_field_path_elements: Dict[str, List[DatabasePathElement]]

                grouped_getters_target_fields_containers[getter_key] = target_field_container
                for item_key, item_target_field_container in target_field_container.items():
                    item_field_object, item_field_path_elements = item_target_field_container
                    found_item_value_in_cache, field_item_value_from_cache = self._cache_get_data(
                        primary_key_value=key_value, field_path_elements=item_field_path_elements
                    )
                    if found_item_value_in_cache is True:
                        if data_validation is True:
                            item_field_object.populate(value=field_item_value_from_cache)
                            validated_data, is_valid = item_field_object.validate_data()
                        else:
                            validated_data = field_item_value_from_cache
                        container_existing_values[item_key] = self.wrap_item_value(item_value=validated_data, from_cache=True)
                    else:
                        container_fields[getter_key] = item_target_field_container
                        getters_database_paths.append(item_field_path_elements)

                grouped_getters_target_fields_containers[getter_key] = container_fields
                existing_values[getter_key] = container_existing_values

        return existing_values, getters_database_paths, single_getters_target_fields_containers, grouped_getters_target_fields_containers

    def _query_multiple_fields(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Tuple[Optional[List[Any]], QueryMetadata]],
            key_value: str, getters: Dict[str, FieldGetter], index_name: Optional[str], data_validation: bool
    ):
        existing_values, getters_database_paths, single_getters_target_fields_containers, grouped_getters_target_fields_containers = (
            self._prepare_getters_with_cache(data_validation=data_validation, key_value=key_value, getters=getters)
        )
        if index_name is None or index_name == self.primary_index_name:
            records_attributes, query_metadata = middleware(getters_database_paths)
            if records_attributes is not None and len(records_attributes) > 0:
                # Since we query the primary_index, we know for a fact that we will never be returned more than
                # one record item. Hence why we do not have a loop that iterate over the records_items_data,
                # and that we return a dict with the only one record item being the requested key_value.
                index_cached_data: dict = self._index_cached_data(primary_key_value=key_value)
                record_values: Dict[str, Any] = self.unpack_validate_cache_getters_record_attributes_if_need_to(
                    single_getters_target_fields_containers=single_getters_target_fields_containers,
                    grouped_getters_target_fields_containers=grouped_getters_target_fields_containers,
                    data_validation=data_validation, record_attributes=records_attributes[0],
                    index_cached_data=index_cached_data, base_output_values=existing_values
                )
                return {key_value: record_values}, query_metadata
            return None, query_metadata
        else:
            records_attributes, query_metadata = self.inner_query_fields_secondary_index(
                middleware=middleware, fields_database_paths=getters_database_paths
            )
            index_cached_data: dict = self._index_cached_data(primary_key_value=key_value)
            output_records_values: dict = {
                record_key: self.unpack_validate_cache_getters_record_attributes_if_need_to(
                    data_validation=data_validation, record_attributes=record_item_attributes, index_cached_data=index_cached_data,
                    single_getters_target_fields_containers=single_getters_target_fields_containers,
                    grouped_getters_target_fields_containers=grouped_getters_target_fields_containers
                ) for record_key, record_item_attributes in records_attributes.items()
            }
            return output_records_values, query_metadata

    def _get_multiple_fields(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Any],
            key_value: str, getters: Dict[str, FieldGetter], data_validation: bool
    ) -> Optional[dict]:
        output_data: Dict[str, Any] = {}

        single_getters_target_fields_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
        grouped_getters_target_fields_containers: Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]] = {}

        getters_database_paths: List[List[DatabasePathElement]] = []
        for getter_key, getter_item in getters.items():
            target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
                field_path=getter_item.field_path, fields_switch=self.fields_switch, query_kwargs=getter_item.query_kwargs
            )
            if is_multi_selector is not True:
                target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                field_object, field_path_elements = target_field_container

                found_value_in_cache, field_value_from_cache = self._cache_get_data(
                    primary_key_value=key_value, field_path_elements=field_path_elements
                )
                if found_value_in_cache is True:
                    output_data[getter_key] = self._validate_format_field_value_if_need_to(
                        value=field_value_from_cache, data_validation=data_validation,
                        field_object=field_object, from_cache=True
                    )
                else:
                    single_getters_target_fields_containers[getter_key] = target_field_container
                    getters_database_paths.append(field_path_elements)
            else:
                target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

                current_getter_grouped_database_paths_elements: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
                container_data: Dict[str, Any] = {}

                for item_key, item_container in target_field_container.items():
                    item_field_object, item_field_path_elements = item_container

                    found_item_value_in_cache, field_item_value_from_cache = self._cache_get_data(
                        primary_key_value=key_value, field_path_elements=item_field_path_elements
                    )
                    if found_item_value_in_cache is True:
                        container_data[item_key] = self._validate_format_field_value_if_need_to(
                            value=key_value, data_validation=data_validation,
                            field_object=item_field_object, from_cache=True
                        )
                    else:
                        current_getter_grouped_database_paths_elements[item_key] = item_container
                        getters_database_paths.append(item_field_path_elements)

                output_data[getter_key] = container_data
                grouped_getters_target_fields_containers[getter_key] = current_getter_grouped_database_paths_elements

        record_attributes: Optional[dict] = middleware(getters_database_paths)
        if record_attributes is None:
            return output_data

        index_cached_data: dict = self._index_cached_data(primary_key_value=key_value)
        unpacked_retrieved_items: dict = self.unpack_validate_cache_getters_record_attributes_if_need_to(
            data_validation=data_validation, record_attributes=record_attributes, index_cached_data=index_cached_data,
            single_getters_target_fields_containers=single_getters_target_fields_containers,
            grouped_getters_target_fields_containers=grouped_getters_target_fields_containers
        )
        return {**output_data, **unpacked_retrieved_items}

    def _update_field(self, key_value: str, field_path: str, value_to_set: Any, query_kwargs: Optional[dict] = None) -> bool:
        field_object, field_path_elements, validated_data, is_valid = process_validate_data_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs, data_to_validate=value_to_set
        )
        if is_valid is True:
            index_cached_data = self._index_cached_data(primary_key_value=key_value)
            BaseCachingTable._cache_put_data(index_cached_data=index_cached_data, field_path_elements=field_path_elements, data=validated_data)

            joined_field_path = join_field_path_elements(field_path_elements)
            pending_update_operations = self._index_pending_update_operations(primary_key_value=key_value)
            pending_update_operations[joined_field_path] = FieldPathSetter(
                field_path_elements=field_path_elements, value_to_set=validated_data
            )
            return True
        return False

    def _update_field_return_old(
            self, middleware: Callable[[List[DatabasePathElement], Any], Tuple[bool, Optional[Any]]],
            key_value: str, field_path: str, value_to_set: Any, query_kwargs: Optional[dict], data_validation: bool
    ) -> Tuple[bool, Optional[Any]]:

        field_object, field_path_elements, validated_update_data, update_data_is_valid = process_validate_data_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs, data_to_validate=value_to_set
        )
        if update_data_is_valid is not True:
            return False, None

        index_cached_data = self._index_cached_data(primary_key_value=key_value)

        field_path_elements: List[DatabasePathElement]
        found_value_in_cache, field_value_from_cache = self._cache_get_data(
            primary_key_value=key_value, field_path_elements=field_path_elements
        )
        if found_value_in_cache is True:
            # When the old value of our field is found in cache, we are not forced to perform a database operation right away in order
            # to know the old value of our field, but we still need to schedule an update operation, then update the in memory cache.
            joined_field_path = join_field_path_elements(field_path_elements)
            pending_update_operations = self._index_pending_update_operations(primary_key_value=key_value)
            pending_update_operations[joined_field_path] = FieldPathSetter(
                field_path_elements=field_path_elements, value_to_set=validated_update_data
            )

            self._cache_put_data(
                index_cached_data=index_cached_data,
                field_path_elements=field_path_elements,
                data=validated_update_data
            )
            return True, self._validate_format_field_value_if_need_to(
                value=field_value_from_cache, data_validation=data_validation,
                field_object=field_object, from_cache=True
            )
        else:
            update_success, response_attributes = middleware(field_path_elements, validated_update_data)
            if update_success is not True:
                return False, None

            self._cache_put_data(
                index_cached_data=index_cached_data,
                field_path_elements=field_path_elements,
                data=validated_update_data
            )

            old_item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
                data=response_attributes, field_path_elements=field_path_elements,
                num_keys_to_navigation_into=len(field_path_elements)
            ) if response_attributes is not None else None

            return update_success, self._validate_format_field_value_if_need_to(
                value=old_item_data, data_validation=data_validation,
                field_object=field_object, from_cache=False
            )

    def _update_multiple_fields(self, key_value: str, setters: List[FieldSetter or UnsafeFieldSetter]) -> bool:
        index_cached_data = self._index_cached_data(primary_key_value=key_value)
        for current_setter in setters:
            if isinstance(current_setter, FieldSetter):
                field_object, field_path_elements, validated_data, is_valid = process_validate_data_and_make_single_rendered_database_path(
                    field_path=current_setter.field_path, fields_switch=self.fields_switch,
                    query_kwargs=current_setter.query_kwargs, data_to_validate=current_setter.value_to_set
                )
                if is_valid is True:
                    BaseCachingTable._cache_put_data(index_cached_data=index_cached_data, field_path_elements=field_path_elements, data=validated_data)
                    joined_field_path = join_field_path_elements(field_path_elements)
                    pending_update_operations = self._index_pending_update_operations(primary_key_value=key_value)
                    pending_update_operations[joined_field_path] = FieldPathSetter(
                        field_path_elements=field_path_elements, value_to_set=validated_data
                    )
            elif isinstance(current_setter, UnsafeFieldSetter):
                raise Exception(f"UnsafeFieldSetter not supported in caching_table")
                """safe_field_path_object, is_multi_selector = process_and_get_field_path_object_from_field_path(
                    field_path_key=current_setter.safe_base_field_path, fields_switch=self.fields_switch
                )
                # todo: add support for multiple fields path
                if current_setter.unsafe_path_continuation is None:
                    field_path_elements = safe_field_path_object.database_path
                else:
                    field_path_elements = safe_field_path_object.database_path + current_setter.unsafe_path_continuation

                processed_value_to_set: Any = float_to_decimal_serializer(current_setter.value_to_set)
                # Since the data is not validated, we pass it to the float_to_decimal_serializer
                # function (which normally should be called by the data validation function)

                rendered_field_path_elements = make_rendered_database_path(
                    database_path_elements=field_path_elements,
                    query_kwargs=current_setter.query_kwargs
                )
                dynamodb_setters.append(FieldPathSetter(
                    field_path_elements=rendered_field_path_elements,
                    value_to_set=processed_value_to_set
                ))"""
        return True

    def _update_multiple_fields_return_old(
            self, middleware: Callable[[Dict[str, FieldPathSetter]], Tuple[bool, Optional[dict]]],
            key_value: str, setters: Dict[str, FieldSetter], data_validation: bool
    ) -> Tuple[bool, Dict[str, Optional[Any]]]:

        setters_containers: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
        dynamodb_setters: Dict[str, FieldPathSetter] = {}
        output_data: Dict[str, Optional[Any]] = {}
        index_cached_data = self._index_cached_data(primary_key_value=key_value)

        for setter_key, setter_item in setters.items():
            field_object, field_path_elements, validated_data, is_valid = process_validate_data_and_make_single_rendered_database_path(
                field_path=setter_item.field_path, fields_switch=self.fields_switch,
                query_kwargs=setter_item.query_kwargs, data_to_validate=setter_item.value_to_set
            )
            if is_valid is True:
                found_value_in_cache, field_value_from_cache = self._cache_get_data(
                    primary_key_value=key_value, field_path_elements=field_path_elements
                )
                if found_value_in_cache is True:
                    # When the old value of our field is found in cache, we are not forced to perform a database operation right away in order
                    # to know the old value of our field, but we still need to schedule an update operation, then update the in memory cache.
                    joined_field_path = join_field_path_elements(field_path_elements)
                    pending_update_operations = self._index_pending_update_operations(primary_key_value=key_value)
                    pending_update_operations[joined_field_path] = FieldPathSetter(
                        field_path_elements=field_path_elements, value_to_set=validated_data
                    )

                    self._cache_put_data(
                        index_cached_data=index_cached_data,
                        field_path_elements=field_path_elements,
                        data=validated_data
                    )

                    output_data[setter_key] = self._validate_format_field_value_if_need_to(
                        value=field_value_from_cache, data_validation=data_validation,
                        field_object=field_object, from_cache=True
                    )
                else:
                    setters_containers[setter_key] = (field_object, field_path_elements)
                    dynamodb_setters[setter_key] = FieldPathSetter(
                        field_path_elements=field_path_elements, value_to_set=validated_data
                    )
                    self._cache_put_data(
                        index_cached_data=index_cached_data,
                        field_path_elements=field_path_elements,
                        data=validated_data
                    )

        if not len(dynamodb_setters) > 0:
            # If all the fields old values have been found in cache, and their update
            # have been scheduled, but we do not need to send a database request.
            return True, output_data

        update_success, setters_response_attributes = middleware(dynamodb_setters)

        for item_key, item_container in setters_containers.items():
            item_field_object, item_field_path_elements = item_container

            item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
                data=setters_response_attributes, field_path_elements=item_field_path_elements,
                num_keys_to_navigation_into=len(item_field_path_elements)
            )
            output_data[item_key] = self._validate_format_field_value_if_need_to(
                value=item_data, data_validation=data_validation,
                field_object=item_field_object, from_cache=False
            )

        return update_success, output_data

    def _remove_field(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Any],
            key_value: str, field_path: str, query_kwargs: Optional[dict], data_validation: bool
    ) -> Optional[Any]:

        target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
            field_path=field_path, fields_switch=self.fields_switch, query_kwargs=query_kwargs
        )
        index_cached_data = self._index_cached_data(primary_key_value=key_value)

        if is_multi_selector is not True:
            target_field_container: Tuple[BaseField, List[DatabasePathElement]]
            field_object, field_path_elements = target_field_container

            found_value_in_cache, field_value_from_cache = self._cache_get_data(
                primary_key_value=key_value, field_path_elements=field_path_elements
            )
            if found_value_in_cache is True:
                pending_remove_operations = self._index_pending_remove_operations(primary_key_value=key_value)
                self._cache_add_delete_operation(
                    index_cached_data=index_cached_data,
                    pending_remove_operations=pending_remove_operations,
                    field_path_elements=field_path_elements
                )
                # Even when we retrieve a removed value from the cache, and that we do not need to perform a remove operation right away to retrieve
                # the removed value, we still want to add a delete_operation that will be performed on operation commits, because if we remove a value
                # from the cache, it does not remove a potential older value present in the database, that the remove operation should remove.
                return self._validate_format_field_value_if_need_to(
                    value=field_value_from_cache, data_validation=data_validation,
                    field_object=field_object, from_cache=True
                )
            else:
                target_path_elements: List[List[DatabasePathElement]] = [field_path_elements]
                self._cache_remove_field(
                    index_cached_data=index_cached_data,
                    primary_key_value=key_value,
                    field_path_elements=field_path_elements
                )
                response_attributes: Optional[dict] = middleware(target_path_elements)
                removed_item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
                    data=response_attributes, field_path_elements=field_path_elements,
                    num_keys_to_navigation_into=len(field_path_elements)
                )
                return self._validate_format_field_value_if_need_to(
                    value=removed_item_data, data_validation=data_validation,
                    field_object=field_object, from_cache=False
                )
        else:
            target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

            target_path_elements: List[List[DatabasePathElement]] = []
            container_output_data: Dict[str, Any] = {}

            for item_key, item_container in target_field_container.items():
                item_field_object, item_field_path_elements = item_container

                found_item_value_in_cache, field_item_value_from_cache = self._cache_get_data(
                    primary_key_value=key_value, field_path_elements=item_field_path_elements
                )
                if found_item_value_in_cache is True:
                    pending_remove_operations = self._index_pending_remove_operations(primary_key_value=key_value)
                    self._cache_add_delete_operation(
                        index_cached_data=index_cached_data,
                        pending_remove_operations=pending_remove_operations,
                        field_path_elements=item_field_path_elements
                    )
                    container_output_data[item_key] = self._validate_format_field_value_if_need_to(
                        value=field_item_value_from_cache, data_validation=data_validation,
                        field_object=item_field_object, from_cache=True
                    )
                else:
                    target_path_elements.append(item_field_path_elements)
                    self._cache_remove_field(
                        index_cached_data=index_cached_data,
                        primary_key_value=key_value,
                        field_path_elements=item_field_path_elements
                    )

            if len(target_path_elements) > 0:
                response_attributes: Optional[dict] = middleware(target_path_elements)
                for item_key, item_container in target_field_container.items():
                    item_field_object, item_field_path_elements = item_container

                    removed_item_data: Optional[Any] = navigate_into_data_with_field_path_elements(
                        data=response_attributes, field_path_elements=item_field_path_elements,
                        num_keys_to_navigation_into=len(item_field_path_elements)
                    )
                    container_output_data[item_key] = self._validate_format_field_value_if_need_to(
                        value=removed_item_data, data_validation=data_validation,
                        field_object=item_field_object, from_cache=False
                    )

            return container_output_data

    def _delete_field(self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None) -> bool:
        index_cached_data = self._index_cached_data(primary_key_value=key_value)
        pending_remove_operations = self._index_pending_remove_operations(primary_key_value=key_value)
        self._cache_process_add_delete_operation(
            index_cached_data=index_cached_data,
            pending_remove_operations=pending_remove_operations,
            field_path=field_path, query_kwargs=query_kwargs
        )
        return True

    def _grouped_remove_multiple_fields(
            self, middleware: Callable[[List[List[DatabasePathElement]]], Any],
            key_value: str, removers: Dict[str, FieldRemover], data_validation: bool
    ) -> Optional[Dict[str, Any]]:
        # todo: do not perform the operation but store it as pending if a matching value exists in the cache
        if not len(removers) > 0:
            # If no remover has been specified, we do not run the database
            # operation, and since no value has been removed, we return None.
            return None
        else:
            output_data: dict = {}
            index_cached_data = self._index_cached_data(primary_key_value=key_value)

            removers_field_paths_elements: Dict[str, Tuple[BaseField, List[DatabasePathElement]]] = {}
            grouped_removers_field_paths_elements: Dict[str, Dict[str, Tuple[BaseField, List[DatabasePathElement]]]] = {}

            # todo: add the remove items present in cache to the output_data instead of only returning the items not removed from cache

            removers_database_paths: List[List[DatabasePathElement]] = []
            for remover_key, remover_item in removers.items():
                target_field_container, is_multi_selector = process_and_make_single_rendered_database_path(
                    field_path=remover_item.field_path, fields_switch=self.fields_switch,
                    query_kwargs=remover_item.query_kwargs
                )
                if is_multi_selector is not True:
                    target_field_container: Tuple[BaseField, List[DatabasePathElement]]
                    field_object, field_path_elements = target_field_container

                    removers_field_paths_elements[remover_key] = target_field_container
                    removers_database_paths.append(field_path_elements)
                    self._cache_remove_field(
                        index_cached_data=index_cached_data,
                        primary_key_value=key_value,
                        field_path_elements=field_path_elements
                    )
                else:
                    target_field_container: Dict[str, Tuple[BaseField, List[DatabasePathElement]]]

                    grouped_removers_field_paths_elements[remover_key] = target_field_container
                    for item_container in target_field_container.values():
                        item_field_object, item_field_path_elements = item_container

                        removers_database_paths.append(item_field_path_elements)
                        self._cache_remove_field(
                            index_cached_data=index_cached_data,
                            primary_key_value=key_value,
                            field_path_elements=item_field_path_elements
                        )

            record_attributes = middleware(removers_database_paths)
            if record_attributes is None:
                return None

            index_cached_data: dict = self._index_cached_data(primary_key_value=key_value)
            return self.unpack_validate_cache_getters_record_attributes_if_need_to(
                data_validation=data_validation, record_attributes=record_attributes, index_cached_data=index_cached_data,
                single_getters_target_fields_containers=removers_field_paths_elements,
                grouped_getters_target_fields_containers=grouped_removers_field_paths_elements
            )

    def _grouped_delete_multiple_fields(self, key_value: str, removers: List[FieldRemover]) -> bool:
        index_cached_data = self._index_cached_data(primary_key_value=key_value)
        pending_remove_operations = self._index_pending_remove_operations(primary_key_value=key_value)

        for current_remover in removers:
            self._cache_process_add_delete_operation(
                index_cached_data=index_cached_data,
                pending_remove_operations=pending_remove_operations,
                field_path=current_remover.field_path,
                query_kwargs=current_remover.query_kwargs
            )
        return True



from typing import List

from StructNoSQL import DynamoDBTableConnectors, PrimaryIndex, GlobalSecondaryIndex
from StructNoSQL.clients_middlewares.dynamodb.backend.dynamodb_core import DynamoDbCoreAdapter


class DynamoDBLowLevelTableClient(DynamoDBTableConnectors):
    def __init__(
            self, table_name: str, region_name: str, primary_index: PrimaryIndex,
            billing_mode: str = DynamoDbCoreAdapter.PAY_PER_REQUEST,
            global_secondary_indexes: List[GlobalSecondaryIndex] = None,
            auto_create_table: bool = True
    ):
        self.__setup_connectors__(
            table_name=table_name, region_name=region_name,
            primary_index=primary_index, global_secondary_indexes=global_secondary_indexes,
            billing_mode=billing_mode, auto_create_table=auto_create_table
        )



from typing import Dict
from StructNoSQL import TableDataModel, BaseField, MapModel


class BaseTableModel(TableDataModel):
    type = BaseField(field_type=str, required=False)
    fieldOne = BaseField(field_type=str, required=False)
    fieldTwo = BaseField(field_type=str, required=False)
    class ContainerModel(MapModel):
        fieldOne = BaseField(field_type=str, required=False)
        fieldTwo = BaseField(field_type=str, required=False)
        fieldThree = BaseField(field_type=str, required=False)
    container = BaseField(field_type=Dict[str, ContainerModel], key_name='containerKey', required=False)

class DynamoDBTableModel(BaseTableModel):
    accountId = BaseField(field_type=str, required=True)

class ExternalDynamoDBApiTableModel(BaseTableModel):
    accountProjectTableKeyId = BaseField(field_type=str, required=True)



import unittest

from tests.components.playground_table_clients import PlaygroundDynamoDBCachingTable
from tests.tests_data_retrieval_validation.table_models import DynamoDBFirstTableModel, DynamoDBSecondTableModel


class TestsDynamoDBCachingTable(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.first_table = PlaygroundDynamoDBCachingTable(data_model=DynamoDBFirstTableModel)
        self.first_table.debug = True
        self.second_table = PlaygroundDynamoDBCachingTable(data_model=DynamoDBSecondTableModel)
        self.second_table.debug = True

        self.SHARED_CASE_KWARGS = {
            'self': self, 'first_table': self.first_table, 'second_table': self.second_table,
            'is_caching': True, 'primary_key_name': 'accountId'
        }

    def test_get_field(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_get_field
        test_get_field(**self.SHARED_CASE_KWARGS)

    def test_get_field_multi_selectors(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_get_field_multi_selectors
        test_get_field_multi_selectors(**self.SHARED_CASE_KWARGS)

    def test_get_multiple_fields(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_get_multiple_fields
        test_get_multiple_fields(**self.SHARED_CASE_KWARGS)

    def test_remove_field(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_remove_field
        test_remove_field(**self.SHARED_CASE_KWARGS)

    def test_remove_field_multi_selectors(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_remove_field_multi_selectors
        test_remove_field_multi_selectors(**self.SHARED_CASE_KWARGS)

    def test_remove_multiple_fields(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_remove_multiple_fields
        test_remove_multiple_fields(**self.SHARED_CASE_KWARGS)

    def test_grouped_remove_multiple_fields(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_grouped_remove_multiple_fields
        test_grouped_remove_multiple_fields(**self.SHARED_CASE_KWARGS)

    def test_update_field_return_old(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_update_field_return_old
        test_update_field_return_old(**self.SHARED_CASE_KWARGS)

    def test_update_multiple_fields_return_old(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_update_multiple_fields_return_old
        test_update_multiple_fields_return_old(**self.SHARED_CASE_KWARGS)

    def test_query_field(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_query_field
        test_query_field(**self.SHARED_CASE_KWARGS)

    def test_query_field_multi_selectors(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_query_field_multi_selectors
        test_query_field_multi_selectors(**self.SHARED_CASE_KWARGS)

    def test_query_multiple_fields(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_query_multiple_fields
        test_query_multiple_fields(**self.SHARED_CASE_KWARGS)

    def test_remove_record(self):
        from tests.tests_data_retrieval_validation.cases_shared import test_remove_record
        test_remove_record(**self.SHARED_CASE_KWARGS)



import unittest

from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable
from tests.tests_remove_record.table_models import DynamoDBTableModel


class TestsDynamoDBBasicTable(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=DynamoDBTableModel)

        self.DYNAMODB_CASE_KWARGS = {'self': self, 'users_table': self.users_table, 'is_caching': False}
        self.SHARED_CASE_KWARGS = {**self.DYNAMODB_CASE_KWARGS, 'primary_key_name': 'accountId'}

    def test_basic_record_removal(self):
        from tests.tests_remove_record.cases_shared import test_basic_record_removal
        test_basic_record_removal(**self.SHARED_CASE_KWARGS)



import unittest
from uuid import uuid4

from StructNoSQL import BaseField, MapModel, NoneType, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    class ItemModel(MapModel):
        value = BaseField(field_type=str, required=True)
    requiredModelAcceptingNone = BaseField(field_type=ItemModel, required=False)
    # This field only accept the ItemModel as field_type, but since the field is not required, will be added to the field_type.


class TestNonRequiredFields(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_required_model_allowing_none(self):
        random_value_one: str = str(uuid4())
        update_one_success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path='requiredModelAcceptingNone',
            value_to_set={'value': random_value_one}
        )
        self.assertTrue(update_one_success)

        update_two_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path='requiredModelAcceptingNone',
            value_to_set="dummyDataThatShouldBeRejected"
        )
        self.assertFalse(update_two_success)

        update_three_success = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path='requiredModelAcceptingNone',
            value_to_set=None
        )
        self.assertTrue(update_three_success)


if __name__ == '__main__':
    unittest.main()



from StructNoSQL import TableDataModel, BaseField


class BaseTableModel(TableDataModel):
    simpleField = BaseField(field_type=str, required=False)
    fieldWithCustomName = BaseField(custom_field_name='field%/!$', field_type=str, required=False)

class DynamoDBTableModel(BaseTableModel):
    accountId = BaseField(field_type=str, required=True)

class ExternalDynamoDBApiTableModel(BaseTableModel):
    accountProjectTableKeyId = BaseField(field_type=str, required=True)



import unittest
from typing import Set, Optional
from uuid import uuid4

from StructNoSQL import BaseField, MapModel, TableDataModel
from StructNoSQL.exceptions import UsageOfUntypedSetException
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    class ContainerModel(MapModel):
        typedSet = BaseField(field_type=Set[str], key_name='setKey', required=False)
    container = BaseField(field_type=ContainerModel)

class TestsSetObjectType(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_crash_on_untyped_set(self):
        def init_table():
            class TableModel(TableDataModel):
                accountId = BaseField(field_type=str, required=True)
                untypedSet = BaseField(field_type=set, key_name='setKey', required=False)
            users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel())
        self.assertRaises(UsageOfUntypedSetException, init_table)

    def test_set_retrieve_individual_typed_set_item(self):
        valid_random_set_values: Set[str] = set(str(uuid4()) for i in range(5))
        single_valid_set_item = list(valid_random_set_values)[2]
        random_set_values = valid_random_set_values.copy()
        random_set_values.add(42)
        # Add invalid value to the random_set_values

        set_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path='container.typedSet', value_to_set=random_set_values
        )
        self.assertTrue(set_success)

        # todo: add support for retrieving single set item (we need to check if the set item exists)
        retrieved_set_item: Optional[str] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID, field_path='container.typedSet.{{setKey}}', query_kwargs={'setKey': "ee"}
        )
        self.assertEqual(retrieved_set_item, single_valid_set_item)

        retrieved_entire_set: Optional[Set[str]] = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID, field_path='container.typedSet'
        )
        self.assertEqual(valid_random_set_values, retrieved_entire_set)

        deletion_success: bool = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID, field_path='container.typedSet'
        )
        self.assertTrue(deletion_success)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, List, Dict, Any, Tuple, Union, Type

from pydantic import BaseModel

from StructNoSQL import PrimaryIndex, TableDataModel
from StructNoSQL.models import DatabasePathElement, FieldGetter, FieldSetter, UnsafeFieldSetter, FieldRemover, \
    FieldPathSetter, QueryMetadata
from StructNoSQL.practical_logger import message_with_vars
from StructNoSQL.tables.base_basic_table import BaseBasicTable
from StructNoSQL.clients_middlewares.external_dynamodb_api.external_dynamodb_api_table_connectors import ExternalDynamoDBApiTableConnectors


class ExternalDynamoDBApiBasicTable(BaseBasicTable, ExternalDynamoDBApiTableConnectors):
    def __init__(
            self, data_model: Type[TableDataModel], primary_index: PrimaryIndex,
            api_http_endpoint: str, base_payload: Optional[dict] = None
    ):
        super().__init__(data_model=data_model, primary_index=primary_index)
        self.__setup_connectors__(api_http_endpoint=api_http_endpoint, base_payload=base_payload)

    def put_record(self, record_dict_data: dict) -> bool:
        def middleware(validated_record_item: dict) -> bool:
            return self._put_record_request(record_item_data=validated_record_item)
        return self._put_record(middleware=middleware, record_dict_data=record_dict_data)

    def delete_record(self, indexes_keys_selectors: dict) -> bool:
        def middleware(indexes_keys: dict) -> bool:
            return self._delete_record_request(indexes_keys_selectors=indexes_keys)
        return self._delete_record(middleware=middleware, indexes_keys_selectors=indexes_keys_selectors)

    def remove_record(self, indexes_keys_selectors: dict, data_validation: bool = True) -> Optional[dict]:
        def middleware(indexes_keys: dict) -> Optional[dict]:
            return self._remove_record_request(indexes_keys_selectors=indexes_keys)
        return self._remove_record(middleware=middleware, indexes_keys_selectors=indexes_keys_selectors, data_validation=data_validation)

    def query_field(
            self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None, pagination_records_limit: Optional[int] = None,
            filter_expression: Optional[Any] = None, data_validation: bool = True, **additional_kwargs
    ) -> Tuple[Optional[dict], QueryMetadata]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]) -> Tuple[Optional[List[dict]], QueryMetadata]:
            return self._query_items_by_key(
                key_value=key_value, fields_path_elements=fields_path_elements,
                pagination_records_limit=pagination_records_limit, filter_expression=filter_expression,
                **additional_kwargs
            )
        return self._query_field(
            middleware=middleware, key_value=key_value, field_path=field_path,
            query_kwargs=query_kwargs, index_name=None, data_validation=data_validation
        )

    def query_multiple_fields(
            self, key_value: str, getters: Dict[str, FieldGetter], pagination_records_limit: Optional[int] = None,
            filter_expression: Optional[Any] = None, data_validation: bool = True, **additional_kwargs
    ) -> Tuple[Optional[Dict[str, Any]], QueryMetadata]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]) -> Tuple[Optional[List[dict]], QueryMetadata]:
            return self._query_items_by_key(
                key_value=key_value, fields_path_elements=fields_path_elements,
                pagination_records_limit=pagination_records_limit,
                filter_expression=filter_expression, **additional_kwargs
            )
        return self._query_multiple_fields(
            middleware=middleware, key_value=key_value, getters=getters,
            index_name=None, data_validation=data_validation
        )

    def get_field(self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None, data_validation: bool = True) -> Any:
        def middleware(field_path_elements: Union[List[DatabasePathElement], Dict[str, List[DatabasePathElement]]], is_multi_selector: bool):
            if is_multi_selector is not True:
                field_path_elements: List[DatabasePathElement]
                return self._get_single_value_in_path_target(key_value=key_value, field_path_elements=field_path_elements)
            else:
                field_path_elements: Dict[str, List[DatabasePathElement]]
                return self._get_values_in_multiple_path_target(key_value=key_value, fields_path_elements=field_path_elements)
        return self._get_field(middleware=middleware, field_path=field_path, query_kwargs=query_kwargs, data_validation=data_validation)

    def get_multiple_fields(self, key_value: str, getters: Dict[str, FieldGetter], data_validation: bool = True) -> Optional[dict]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]):
            return self._get_or_query_single_item(
                key_value=key_value, fields_path_elements=fields_path_elements,
            )
        return self._get_multiple_fields(middleware=middleware, getters=getters, data_validation=data_validation)
        
    # todo: implement query_field and query_multiple_fields

    def update_field(self, key_value: str, field_path: str, value_to_set: Any, query_kwargs: Optional[dict] = None) -> bool:
        def middleware(field_path_elements: List[DatabasePathElement], validated_data: Any):
            response = self._set_update_data_element_to_map_with_default_initialization(
                key_value=key_value, value=validated_data,
                field_path_elements=field_path_elements
            )
            return True if response is not None else False
        return self._update_field(middleware=middleware, field_path=field_path, value_to_set=value_to_set, query_kwargs=query_kwargs)

    def update_field_return_old(
            self, key_value: str, field_path: str, value_to_set: Any,
            query_kwargs: Optional[dict] = None, data_validation: bool = True
    ) -> Tuple[bool, Optional[Any]]:
        def middleware(field_path_elements: List[DatabasePathElement], validated_data: Any) -> Tuple[bool, Optional[Any]]:
            update_success, response_attributes = self._set_update_data_element_to_map_with_default_initialization_return_old(
                key_value=key_value, value=validated_data,
                field_path_elements=field_path_elements
            )
            return update_success, response_attributes
        return self._update_field_return_old(
            middleware=middleware, field_path=field_path, value_to_set=value_to_set,
            query_kwargs=query_kwargs, data_validation=data_validation
        )

    def update_multiple_fields(self, key_value: str, setters: List[FieldSetter or UnsafeFieldSetter]) -> bool:
        def middleware(dynamodb_setters: List[FieldPathSetter]) -> bool:
            update_success: bool = self._set_update_multiple_data_elements_to_map(
                key_value=key_value, setters=dynamodb_setters
            )
            return update_success
        return self._update_multiple_fields(middleware=middleware, setters=setters)

    def update_multiple_fields_return_old(self, key_value: str, setters: Dict[str, FieldSetter], data_validation: bool = True) -> Tuple[bool, Dict[str, Optional[Any]]]:
        def middleware(dynamodb_setters: Dict[str, FieldPathSetter]) -> Tuple[bool, dict]:
            update_success, response_attributes = self._set_update_multiple_data_elements_to_map_return_old(
                key_value=key_value, setters=list(dynamodb_setters.values())
            )
            return update_success, response_attributes
        return self._update_multiple_fields_return_old(middleware=middleware, setters=setters, data_validation=data_validation)

    def remove_field(self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None, data_validation: bool = True) -> Optional[Any]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]) -> Optional[dict]:
            return self._remove_data_elements_from_map(
                key_value=key_value, fields_path_elements=fields_path_elements,
            )
        return self._remove_field(middleware=middleware, field_path=field_path, query_kwargs=query_kwargs, data_validation=data_validation)

    def remove_multiple_fields(self, key_value: str, removers: Dict[str, FieldRemover], data_validation: bool = True) -> Dict[str, Optional[Any]]:
        def task_executor(remover_item: FieldRemover):
            return self.remove_field(
                key_value=key_value,
                field_path=remover_item.field_path,
                query_kwargs=remover_item.query_kwargs,
                data_validation=data_validation
            )
        return self._async_field_removers_executor(task_executor=task_executor, removers=removers)

    def delete_field(self, key_value: str, field_path: str, query_kwargs: Optional[dict] = None) -> bool:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]) -> bool:
            return self._remove_data_elements_from_map(key_value=key_value, fields_path_elements=fields_path_elements)
        return self._delete_field(middleware=middleware, field_path=field_path, query_kwargs=query_kwargs)

    def delete_multiple_fields(self, key_value: str, removers: Dict[str, FieldRemover]) -> Dict[str, bool]:
        def task_executor(remover_item: FieldRemover):
            return self.delete_field(
                key_value=key_value,
                field_path=remover_item.field_path,
                query_kwargs=remover_item.query_kwargs
            )
        return self._async_field_removers_executor(task_executor=task_executor, removers=removers)

    def grouped_remove_multiple_fields(self, key_value: str, removers: Dict[str, FieldRemover], data_validation: bool = True) -> Optional[Dict[str, Any]]:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]):
            return self._remove_data_elements_from_map(
                key_value=key_value, fields_path_elements=fields_path_elements,
            )
        return self._grouped_remove_multiple_fields(middleware=middleware, removers=removers, data_validation=data_validation)

    def grouped_delete_multiple_fields(self, key_value: str, removers: List[FieldRemover]) -> bool:
        def middleware(fields_path_elements: List[List[DatabasePathElement]]):
            return self._delete_data_elements_from_map(
                key_value=key_value, fields_path_elements=fields_path_elements,
            )
        return self._grouped_delete_multiple_fields(middleware=middleware, removers=removers)




from StructNoSQL import TableDataModel, BaseField, MapModel


class BaseTableModel(TableDataModel):
    simpleTextField = BaseField(field_type=str, required=False)
    class ContainerModel(MapModel):
        textFieldOne = BaseField(field_type=str, required=False)
        textFieldTwo = BaseField(field_type=str, required=False)
    container = BaseField(field_type=ContainerModel, required=False)

class DynamoDBTableModel(BaseTableModel):
    accountId = BaseField(field_type=str, required=True)

class ExternalDynamoDBApiTableModel(BaseTableModel):
    accountProjectTableKeyId = BaseField(field_type=str, required=True)



import random
import unittest

from StructNoSQL import FieldSetter
from tests.components.playground_table_clients import PlaygroundDynamoDBCachingTable
from tests.tests_caching_table.table_models import DynamoDBTableModel
from tests.components.playground_table_clients import TEST_ACCOUNT_ID, TEST_ACCOUNT_USERNAME


class TestDynamoDBCachingTable(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBCachingTable(data_model=DynamoDBTableModel)
        self.users_table.debug = True

    def test_simple_get_field(self):
        from tests.tests_caching_table.cases_shared import test_simple_get_field
        test_simple_get_field(self, users_table=self.users_table)

    def test_set_then_get_field_from_cache(self):
        from tests.tests_caching_table.cases_shared import test_set_then_get_field_from_cache
        test_set_then_get_field_from_cache(self, users_table=self.users_table)

    def test_set_then_get_multiple_fields(self):
        from tests.tests_caching_table.cases_shared import test_set_then_get_multiple_fields
        test_set_then_get_multiple_fields(self, users_table=self.users_table)

    def test_set_then_get_pack_values_with_one_of_them_present_in_cache(self):
        from tests.tests_caching_table.cases_shared import test_set_then_get_pack_values_with_one_of_them_present_in_cache
        test_set_then_get_pack_values_with_one_of_them_present_in_cache(self, users_table=self.users_table)

    def test_debug_simple_set_commit_then_get_field_from_database(self):
        from tests.tests_caching_table.cases_shared import test_debug_simple_set_commit_then_get_field_from_database
        test_debug_simple_set_commit_then_get_field_from_database(self, users_table=self.users_table)

    def test_update_multiple_fields(self):
        from tests.tests_caching_table.cases_shared import test_update_multiple_fields
        test_update_multiple_fields(self, users_table=self.users_table)

    def test_set_delete_field(self):
        from tests.tests_caching_table.cases_shared import test_set_delete_field
        test_set_delete_field(self, users_table=self.users_table)

    def test_set_remove_field(self):
        from tests.tests_caching_table.cases_shared import test_set_remove_field
        test_set_remove_field(self, users_table=self.users_table)

    def test_dict_data_unpacking(self):
        from tests.tests_caching_table.cases_shared import test_dict_data_unpacking
        test_dict_data_unpacking(self, users_table=self.users_table)

    def test_list_data_unpacking(self):
        from tests.tests_caching_table.cases_shared import test_list_data_unpacking
        test_list_data_unpacking(self, users_table=self.users_table)

    def test_set_remove_multi_selector_field_and_field_unpacking(self):
        from tests.tests_caching_table.cases_shared import test_set_remove_multi_selector_field_and_field_unpacking
        test_set_remove_multi_selector_field_and_field_unpacking(self, users_table=self.users_table)

    def test_set_delete_multiple_fields(self):
        from tests.tests_caching_table.cases_shared import test_set_delete_multiple_fields
        test_set_delete_multiple_fields(self, users_table=self.users_table)

    def test_set_remove_multiple_fields(self):
        from tests.tests_caching_table.cases_shared import test_set_remove_multiple_fields
        test_set_remove_multiple_fields(self, users_table=self.users_table)

    def test_set_get_fields_with_secondary_index(self):
        self.users_table.clear_cached_data_and_pending_operations()
        random_field_value_one = random.randint(0, 100)
        random_field_value_two = random.randint(100, 200)

        set_update_success: bool = self.users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='simpleValue', value_to_set=random_field_value_one),
            FieldSetter(field_path='simpleValue2', value_to_set=random_field_value_two)
        ])
        self.assertTrue(set_update_success)

        update_commit_success: bool = self.users_table.commit_operations()
        self.assertTrue(update_commit_success)

        retrieved_values = self.users_table.get_field(key_value=TEST_ACCOUNT_ID, field_path='(simpleValue, simpleValue2)')
        self.assertEqual(retrieved_values, {
            'simpleValue': {'fromCache': True, 'value': random_field_value_one},
            'simpleValue2': {'fromCache': True, 'value': random_field_value_two}
        })

        single_field_not_primary_key, query_metadata = self.users_table.query_field(key_value=TEST_ACCOUNT_USERNAME, index_name='username', field_path='simpleValue')
        self.assertEqual({TEST_ACCOUNT_ID: {'fromCache': False, 'value': random_field_value_one}}, single_field_not_primary_key)

        single_field_primary_key, query_metadata = self.users_table.query_field(key_value=TEST_ACCOUNT_USERNAME, index_name='username', field_path='accountId')
        self.assertEqual({TEST_ACCOUNT_ID: {'fromCache': False, 'value': TEST_ACCOUNT_ID}}, single_field_primary_key)

        multiple_fields_without_primary_key, query_metadata = self.users_table.query_field(
            key_value=TEST_ACCOUNT_USERNAME, index_name='username', field_path='(simpleValue, simpleValue2)'
        )
        self.assertEqual({
            TEST_ACCOUNT_ID: {
                'simpleValue': {'fromCache': False, 'value': random_field_value_one},
                'simpleValue2': {'fromCache': False, 'value': random_field_value_two}
            }}, multiple_fields_without_primary_key
        )

        multiple_fields_with_primary_key, query_metadata = self.users_table.query_field(
            key_value=TEST_ACCOUNT_USERNAME, index_name='username',
            field_path='(accountId, simpleValue, simpleValue2)'
        )
        self.assertEqual({
            TEST_ACCOUNT_ID: {
                'accountId': {'fromCache': False, 'value': TEST_ACCOUNT_ID},
                'simpleValue': {'fromCache': False, 'value': random_field_value_one},
                'simpleValue2': {'fromCache': False, 'value': random_field_value_two}
            }}, multiple_fields_with_primary_key
        )


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import List, Optional, Dict
from uuid import uuid4

from StructNoSQL import FieldGetter, FieldSetter, FieldRemover, BaseField, MapModel, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    fieldToRemove = BaseField(field_type=str, required=False)
    class ContainerModel(MapModel):
        firstNestedValue = BaseField(field_type=str, required=False)
        secondNestedValue = BaseField(field_type=str, required=False)
        thirdNestedValue = BaseField(field_type=str, required=False)
    sophisticatedFieldToRemove = BaseField(field_type=Dict[str, ContainerModel], key_name='id', required=False)

class TestRemoveFieldOperations(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_delete_basic_item_from_path_target(self):
        random_value = f"fieldToRemove_{uuid4()}"

        success_field_set: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='fieldToRemove',
            value_to_set=random_value
        )
        self.assertTrue(success_field_set)

        retrieved_field, query_metadata = self.users_table.query_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='fieldToRemove',
        )
        retrieved_field: Dict[str, dict]
        self.assertEqual(retrieved_field, {TEST_ACCOUNT_ID: random_value})

        retrieved_fields, query_metadata = self.users_table.query_multiple_fields(
            key_value=TEST_ACCOUNT_ID, getters={
                'fieldToRemove': FieldGetter(field_path='fieldToRemove')
            }
        )
        retrieved_fields: Dict[str, dict]
        self.assertEqual(retrieved_fields, {TEST_ACCOUNT_ID: {'fieldToRemove': random_value}})

        """success_field_remove = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove'
        )
        self.assertTrue(success_field_remove)

        retrieved_field_data = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID, field_path='fieldToRemove'
        )
        self.assertIsNone(retrieved_field_data)"""

    """def test_remove_sophisticated_item_from_path_target(self):
        query_kwargs = {'id': 'sampleId'}

        first_message = f"one_{uuid4()}"
        second_message = f"two_{uuid4()}"

        def set_data():
            success_field_set = self.users_table.update_multiple_fields(
                key_value=TEST_ACCOUNT_ID, setters=[
                    FieldSetter(
                        field_path="sophisticatedFieldToRemove.{{id}}.firstNestedValue",
                        query_kwargs=query_kwargs, value_to_set=first_message
                    ),
                    FieldSetter(
                        field_path="sophisticatedFieldToRemove.{{id}}.secondNestedValue",
                        query_kwargs=query_kwargs, value_to_set=second_message
                    )
                ]
            )
            self.assertTrue(success_field_set)
        set_data()

        removed_first_message = self.users_table.remove_field(
            field_path="sophisticatedFieldToRemove.{{id}}.firstNestedValue",
            key_value=TEST_ACCOUNT_ID, query_kwargs=query_kwargs,
        )
        self.assertEqual(first_message, removed_first_message)

        removed_second_message = self.users_table.remove_field(
            field_path="sophisticatedFieldToRemove.{{id}}.secondNestedValue",
            key_value=TEST_ACCOUNT_ID, query_kwargs=query_kwargs
        )
        self.assertEqual(second_message, removed_second_message)

        set_data()
        # Since we removed the individuals items, before we can retrieve
        # the entire data, we need to re-set it in the database.
        removed_entire_item = self.users_table.remove_field(
            field_path="sophisticatedFieldToRemove.{{id}}",
            key_value=TEST_ACCOUNT_ID, query_kwargs=query_kwargs,
        )
        self.assertEqual(removed_entire_item, {'firstNestedValue': first_message, 'secondNestedValue': second_message})
        # Python does not care about the dict ordering when doing a dict comparison

    def test_remove_multiple_fields(self):
        random_id = str(uuid4())
        random_value_one = f"one_{uuid4()}"
        random_value_two = f"two_{uuid4()}"

        update_success = self.users_table.update_multiple_fields(key_value=TEST_ACCOUNT_ID, setters=[
            FieldSetter(field_path='fieldToRemove', value_to_set=random_value_one),
            FieldSetter(
                field_path='sophisticatedFieldToRemove.{{id}}.firstNestedValue',
                query_kwargs={'id': random_id}, value_to_set=random_value_two
            )
        ])
        self.assertTrue(update_success)

        get_response_data = self.users_table.get_multiple_fields(key_value=TEST_ACCOUNT_ID, getters={
            'one': FieldGetter(field_path='fieldToRemove'),
            'two': FieldGetter(field_path='sophisticatedFieldToRemove.{{id}}.firstNestedValue', query_kwargs={'id': random_id})
        })
        self.assertEqual(get_response_data.get('one', None), random_value_one)
        self.assertEqual(get_response_data.get('two', None), random_value_two)

        remove_response_data = self.users_table.remove_multiple_fields(key_value=TEST_ACCOUNT_ID, removers={
            'one': FieldRemover(field_path='fieldToRemove'),
            'two': FieldRemover(field_path='sophisticatedFieldToRemove.{{id}}.firstNestedValue', query_kwargs={'id': random_id})
        })
        print(remove_response_data)
    """

if __name__ == '__main__':
    unittest.main()



import unittest
from typing import Optional, Dict
from uuid import uuid4

from StructNoSQL import BaseField, MapModel, TableDataModel
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    accountId = BaseField(field_type=str, required=True)
    class ContainerItem(MapModel):
        value = BaseField(field_type=int, required=False)
    container = BaseField(field_type=Dict[str, ContainerItem], key_name='itemId', required=False)


class TestFieldKeyWithSpecialChars(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_field_key_with_points(self):
        random_item_id = f".pointBefore.{str(uuid4())}.pointAfter."
        query_kwargs = {'itemId': random_item_id}

        set_success: bool = self.users_table.update_field(
            key_value=TEST_ACCOUNT_ID, field_path='container.{{itemId}}',
            query_kwargs=query_kwargs, value_to_set={'value': 42}
        )
        self.assertTrue(set_success)

        retrieved_value = self.users_table.get_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.{{itemId}}.value',
            query_kwargs=query_kwargs
        )
        self.assertEqual(retrieved_value, 42)

        deletion_success: bool = self.users_table.delete_field(
            key_value=TEST_ACCOUNT_ID,
            field_path='container.{{itemId}}',
            query_kwargs=query_kwargs
        )
        self.assertTrue(deletion_success)


if __name__ == '__main__':
    unittest.main()



import unittest
from uuid import uuid4

from StructNoSQL import BaseField, MapModel, TableDataModel, FieldSetter
from tests.components.playground_table_clients import PlaygroundDynamoDBBasicTable, TEST_ACCOUNT_ID


class TableModel(TableDataModel):
    class FieldsToDeleteContainerModel(MapModel):
        existingItemOneToDelete = BaseField(field_type=str, required=False)
        existingItemTwoToDelete = BaseField(field_type=str, required=False)
    fieldsToDeleteContainer = BaseField(field_type=FieldsToDeleteContainerModel, required=False)

class TestDeleteNonExistingItems(unittest.TestCase):
    def __init__(self, method_name: str):
        super().__init__(methodName=method_name)
        self.users_table = PlaygroundDynamoDBBasicTable(data_model=TableModel)

    def test_delete_nested(self):
        success_delete_item_one = self.users_table.delete_field(key_value=TEST_ACCOUNT_ID, field_path='fieldsToDeleteContainer.existingItemOneToDelete')
        success_delete_item_two = self.users_table.delete_field(key_value=TEST_ACCOUNT_ID, field_path='fieldsToDeleteContainer.existingItemOneToDelete')
        self.assertTrue(success_delete_item_one)
        self.assertTrue(success_delete_item_two)

        random_item_one_value = f"itemOne_{uuid4()}"
        random_item_two_value = f"itemTwo_{uuid4()}"

        update_success = self.users_table.update_multiple_fields(
            key_value=TEST_ACCOUNT_ID, setters=[
                FieldSetter(field_path='fieldsToDeleteContainer.existingItemOneToDelete', value_to_set=random_item_one_value),
                FieldSetter(field_path='fieldsToDeleteContainer.existingItemOneToDelete', value_to_set=random_item_two_value)
            ]
        )
        self.assertTrue(update_success)


if __name__ == '__main__':
    unittest.main()
